# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles classes/named tuples for holding class, field, and attribute value(s)"""

from collections import namedtuple
from typing import Type

from attr import Attribute


class ConfigSpace:
    """Class that holds information about the final values of spock class attributes

    Attributes:
        spock_cls: reference to spock class to store information
        fields: dictionary of the current value of attributes

    """

    def __init__(self, spock_cls: Type, fields: dict):
        """Init call for ConfigSpace class

        Args:
            spock_cls: reference to spock class to store information
            fields: dictionary of the current value of attributes

        """
        self.spock_cls = spock_cls
        self.fields = fields

    @property
    def name(self) -> str:
        """Returns the name of the spock class associated with ConfigSpace"""
        return self.spock_cls.__name__


class AttributeSpace:
    """Class that holds information about a single attribute that is mapped to a ConfigSpace

    Attributes:
        config_space: ConfigSpace that the attribute is contained in
        attribute: current Attribute class

    """

    def __init__(self, attribute: Type[Attribute], config_space: ConfigSpace):
        """Init call for AttributeSpace class

        Args:
            config_space: ConfigSpace that the attribute is contained in
            attribute: current Attribute class
        """
        self.config_space = config_space
        self.attribute = attribute
        self._annotations = None
        self.crypto = False

    @property
    def annotations(self):
        return self._annotations

    @annotations.setter
    def annotations(self, x):
        self._annotations = x

    @property
    def field(self):
        """Returns the field value from the ConfigSpace based on the attribute name"""
        return self.config_space.fields[self.attribute.name]

    @field.setter
    def field(self, value):
        """Setter for the field value from the ConfigSpace based on the attribute name"""
        if isinstance(self.attribute.name, str):
            self.config_space.fields[self.attribute.name] = value
        else:
            raise ValueError


BuilderSpace = namedtuple("BuilderSpace", ["arguments", "spock_space"])



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles all S3 related ops -- allows for s3 functionality to be optional to keep req deps light"""

import attr

try:
    import boto3
    from botocore.client import BaseClient
except ImportError:
    print(
        "Missing libraries to support S3 functionality. Please re-install spock with the extra s3 dependencies -- "
        "pip install spock-config[s3]"
    )
import os
import sys
import typing
from urllib.parse import urlparse

from hurry.filesize import size

from spock.addons.s3.configs import S3Config, S3DownloadConfig, S3UploadConfig


def handle_s3_load_path(path: str, s3_config: S3Config) -> str:
    """Handles loading from S3 uri

    Handles downloading file from a given s3 uri to a local temp location and passing the path back to the handler
    load call

    Args:
        path: s3 uri path
        s3_config: s3_config object

    Returns:
        temp_path: the temporary path of the config file downloaded from s3

    """
    if s3_config is None:
        raise ValueError(
            "Load from S3 -- Missing S3Config object which is necessary to handle S3 style paths"
        )
    bucket, obj, fid = get_s3_bucket_object_name(s3_path=path)
    # Construct the full temp path
    temp_path = f"{s3_config.temp_folder}/{fid}"
    # Strip double slashes if exist
    temp_path = temp_path.replace(r"//", r"/")
    temp_path = download_s3(
        bucket=bucket,
        obj=obj,
        temp_path=temp_path,
        s3_session=s3_config.s3_session,
        download_config=s3_config.download_config,
    )
    return temp_path


def handle_s3_save_path(temp_path: str, s3_path: str, name: str, s3_config: S3Config):
    """Handles saving to S3 uri

    Points to the local spock configuration file and handles getting it up to S3

    Args:
        temp_path: the temporary path the spock config was written out to locally
        s3_path: base s3 uri
        name: spock generated filename
        s3_config: s3_config object

    Returns:
    """
    if s3_config is None:
        raise ValueError(
            "Save to S3 -- Missing S3Config object which is necessary to handle S3 style paths"
        )
    # Fix posix strip
    s3_path = s3_path.replace("s3:/", "s3://")
    bucket, obj, fid = get_s3_bucket_object_name(f"{s3_path}/{name}")
    upload_s3(
        bucket=bucket,
        obj=obj,
        temp_path=temp_path,
        s3_session=s3_config.s3_session,
        upload_config=s3_config.upload_config,
    )


def get_s3_bucket_object_name(s3_path: str) -> typing.Tuple[str, str, str]:
    """Splits a S3 uri into bucket, object, name

    Args:
        s3_path: s3 uri

    Returns:
        bucket
        object
        name

    """
    parsed = urlparse(s3_path)
    return parsed.netloc, parsed.path.lstrip("/"), os.path.basename(parsed.path)


def download_s3(
    bucket: str,
    obj: str,
    temp_path: str,
    s3_session: BaseClient,
    download_config: S3DownloadConfig,
) -> str:
    """Attempts to download the file from the S3 uri to a temp location using any extra arguments to the download

    Args:
        bucket: s3 bucket
        obj: s3 object
        temp_path: local temporary path to write file
        s3_session: current s3 session
        download_config: S3DownloadConfig with extra options for the file transfer

    Returns:
        temp_path: the temporary path of the config file downloaded from s3

    """
    try:
        # Unroll the extra options for those values that are not None
        extra_options = {
            k: v for k, v in attr.asdict(download_config).items() if v is not None
        }
        file_size = s3_session.head_object(Bucket=bucket, Key=obj, **extra_options)[
            "ContentLength"
        ]
        print(f"Attempting to download s3://{bucket}/{obj} (size: {size(file_size)})")
        current_progress = 0
        n_ticks = 50

        def _s3_progress_bar(chunk):
            nonlocal current_progress
            # Increment progress
            current_progress += chunk
            done = int(n_ticks * (current_progress / file_size))
            sys.stdout.write(
                f"\r[%s%s] "
                f"{int(current_progress/file_size) * 100}%%"
                % ("=" * done, " " * (n_ticks - done))
            )
            sys.stdout.flush()
            sys.stdout.write("\n\n")

        # Download with the progress callback
        s3_session.download_file(
            bucket, obj, temp_path, Callback=_s3_progress_bar, ExtraArgs=extra_options
        )
        return temp_path
    except IOError:
        print(
            f"Failed to download file from S3 "
            f"(bucket: {bucket}, object: {obj}) "
            f"and write to {temp_path}"
        )


def upload_s3(
    bucket: str,
    obj: str,
    temp_path: str,
    s3_session: BaseClient,
    upload_config: S3UploadConfig,
):
    """Attempts to upload the local file to the S3 uri using any extra arguments to the upload

    Args:
        bucket: s3 bucket
        obj: s3 object
        temp_path: temporary path of the config file
        s3_session: current s3 session
        upload_config: S3UploadConfig with extra options for the file transfer

    Returns:
    """
    try:
        # Unroll the extra options for those values that are not None
        extra_options = {
            k: v for k, v in attr.asdict(upload_config).items() if v is not None
        }
        file_size = os.path.getsize(temp_path)
        print(f"Attempting to upload s3://{bucket}/{obj} (size: {size(file_size)})")
        current_progress = 0
        n_ticks = 50

        def _s3_progress_bar(chunk):
            nonlocal current_progress
            # Increment progress
            current_progress += chunk
            done = int(n_ticks * (current_progress / file_size))
            sys.stdout.write(
                f"\r[%s%s] "
                f"{int(current_progress/file_size) * 100}%%"
                % ("=" * done, " " * (n_ticks - done))
            )
            sys.stdout.flush()
            sys.stdout.write("\n\n")

        # Upload with progress callback
        s3_session.upload_file(
            temp_path, bucket, obj, Callback=_s3_progress_bar, ExtraArgs=extra_options
        )
    except IOError:
        print(
            f"Failed to upload file to S3 "
            f"(bucket: {bucket}, object: {obj}) "
            f"from {temp_path}"
        )



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles all ops for assembling and pretty printing help info"""

import re
import sys
from enum import EnumMeta
from typing import Callable, Union

from attr import NOTHING


def find_attribute_idx(newline_split_docs):
    """Finds the possible split between the header and Attribute annotations

    Args:
        newline_split_docs: new line split text

    Returns:
        idx: -1 if none or the idx of Attributes

    """
    for idx, val in enumerate(newline_split_docs):
        re_check = re.search(r"(?i)Attribute?s?:", val)
        if re_check is not None:
            return idx
    return -1


def split_docs(obj):
    """Possibly splits head class doc string from attribute docstrings

    Attempts to find the first contiguous line within the Google style docstring to use as the class docstring.
    Splits the docs base on the Attributes tag if present.

    Args:
        obj: class object to rip info from

    Returns:
        class_doc: class docstring if present or blank str
        attr_doc: list of attribute doc strings

    """
    if obj.__doc__ is not None:
        # Split by new line
        newline_split_docs = obj.__doc__.split("\n")
        # Cleanup l/t whitespace
        newline_split_docs = [val.strip() for val in newline_split_docs]
    else:
        newline_split_docs = []
    # Find the break between the class docs and the Attribute section -- if this returns -1 then there is no
    # Attributes section
    attr_idx = find_attribute_idx(newline_split_docs)
    head_docs = newline_split_docs[:attr_idx] if attr_idx != -1 else newline_split_docs
    attr_docs = newline_split_docs[attr_idx:] if attr_idx != -1 else []
    # Grab only the first contiguous line as everything else will probably be too verbose (e.g. the
    # mid-level docstring that has detailed descriptions
    class_doc = ""
    for idx, val in enumerate(head_docs):
        class_doc += f" {val}"
        if idx + 1 != len(head_docs) and head_docs[idx + 1] == "":
            break
    # Clean up any l/t whitespace
    class_doc = class_doc.strip()
    if len(class_doc) > 0:
        class_doc = f"-- {class_doc}"
    return class_doc, attr_docs


def match_attribute_docs(attr_name, attr_docs, attr_type_str, attr_default=NOTHING):
    """Matches class attributes with attribute docstrings via regex

    Args:
        attr_name: attribute name
        attr_docs: list of attribute docstrings
        attr_type_str: str representation of the attribute type
        attr_default: str representation of a possible default value

    Returns:
        dictionary of packed attribute information

    """
    # Regex match each value
    a_str = None
    for a_doc in attr_docs:
        match_re = re.search(r"(?i)^" + attr_name + "?:", a_doc)
        # Find only the first match -- if more than one than ignore
        if match_re:
            a_str = a_doc[match_re.end() :].strip()
    return {
        attr_name: {
            "type": attr_type_str,
            "desc": a_str if a_str is not None else "",
            "default": "(default: " + repr(attr_default) + ")"
            if type(attr_default).__name__ != "_Nothing"
            else "",
            "len": {"name": len(attr_name), "type": len(attr_type_str)},
        }
    }


def handle_attributes_print(info_dict, max_indent: int):
    """Prints attribute information in an argparser style format

    Args:
        info_dict: packed attribute info dictionary to print
        max_indent: max indent for pretty print of help

    """
    # Figure out indents
    max_param_length = max([len(k) for k in info_dict.keys()])
    max_type_length = max([v["len"]["type"] for v in info_dict.values()])
    # Print akin to the argparser
    for k, v in info_dict.items():
        print(
            f"    {k}"
            + (" " * (max_param_length - v["len"]["name"] + max_indent))
            + f'{v["type"]}'
            + (" " * (max_type_length - v["len"]["type"] + max_indent))
            + f'{v["desc"]} {v["default"]}'
        )
    # Blank for spacing :-/
    print("")


def get_type_string(val, nested_others):
    """Gets the type of the attr val as a string

    Args:
        val: current attr being processed
        nested_others: list of nested others to deal with that might have module path info in the string

    Returns:
        type_string: type of the attr as a str

    """
    # Grab the base or type info depending on what is provided
    if "type" in val.metadata:
        type_string = repr(val.metadata["type"])
    elif "base" in val.metadata:
        type_string = val.metadata["base"]
    elif hasattr(val.type, "__name__"):
        type_string = val.type.__name__
    else:
        type_string = str(val.type)
    # Regex out the typing info if present
    type_string = re.sub(r"typing.", "", type_string)
    # Regex out any nested_others that have module path information
    for other_val in nested_others:
        split_other = f"{'.'.join(other_val.split('.')[:-1])}."
        type_string = re.sub(split_other, "", type_string)
    # Regex the string to see if it matches any Enums in the __main__ module space
    # Construct the type with the metadata
    if "optional" in val.metadata:
        type_string = f"Optional[{type_string}]"
    return type_string


def get_from_sys_modules(cls_name):
    """Gets the class from a dot notation name

    Args:
        cls_name: dot notation enum name

    Returns:
        module: enum class

    """
    # Split on dot notation
    split_string = cls_name.split(".")
    module = None
    for idx, val in enumerate(split_string):
        # idx = 0 will always be a call to the sys.modules dict
        if idx == 0:
            module = sys.modules[val]
        # all other idx are paths along the module that need to be traversed
        # idx = -1 will always be the final Enum object name we want to grab (final getattr call)
        else:
            module = getattr(module, val)
    return module


def handle_help_main(
    input_classes: list, module_name: str, extract_fnc: Callable, max_indent: int
):
    """Handles the print of the main class types

    Args:
        input_classes: current set of input classes
        module_name: module name to match
        extract_fnc: function that gets the nested lists within classes
        max_indent: max indent for pretty print of help

    Returns:
        other_list: extended list of other classes/enums to process

    """
    # List to catch Enums and classes and handle post spock wrapped attr classes
    other_list = []
    covered_set = set()
    for attrs_class in input_classes:
        # Split the docs into class docs and any attribute docs
        class_doc, attr_docs = split_docs(attrs_class)
        print("  " + attrs_class.__name__ + f" {class_doc}")
        # Keep a running info_dict of all the attribute level info
        info_dict = {}
        for val in attrs_class.__attrs_attrs__:
            # If the type is an enum we need to handle it outside of this attr loop
            # Match the style of nested enums and return a string of module.name notation
            if isinstance(val.type, EnumMeta):
                other_list.append(f"{val.type.__module__}.{val.type.__name__}")
            # if there is a type (implied Iterable) -- check it for nested Enums or classes
            nested_others = extract_fnc(val, module_name)
            if len(nested_others) > 0:
                other_list.extend(nested_others)
            # Get the type represented as a string
            type_string = get_type_string(val, nested_others)
            info_dict.update(
                match_attribute_docs(val.name, attr_docs, type_string, val.default)
            )
        # Add to covered so we don't print help twice in the case of some recursive nesting
        covered_set.add(f"{attrs_class.__module__}.{attrs_class.__name__}")
        handle_attributes_print(info_dict=info_dict, max_indent=max_indent)
    # Convert the enum list to a set to remove dupes and then back to a list so it is iterable -- set diff to not
    # repeat
    return list(set(other_list) - covered_set)


def handle_help_enums(
    other_list: list, module_name: str, extract_fnc: Callable, max_indent: int
):
    """Handles any extra enums from non main args

    Args:
        other_list: extended list of other classes/enums to process
        module_name: module name to match
        extract_fnc: function that gets the nested lists within classes
        max_indent: max indent for pretty print of help

    Returns:
        None

    """
    # Iterate any Enum type classes
    for other in other_list:
        # if it's longer than 2 then it's an embedded Spock class
        if ".".join(other.split(".")[:-1]) == module_name:
            class_type = get_from_sys_modules(other)
            # Invoke recursive call for the class
            attrs_help(
                [class_type],
                module_name,
                extract_fnc=extract_fnc,
                max_indent=max_indent,
            )
        # Fall back to enum style
        else:
            enum = get_from_sys_modules(other)
            # Split the docs into class docs and any attribute docs
            class_doc, attr_docs = split_docs(enum)
            print("  " + enum.__name__ + f" ({class_doc})")
            info_dict = {}
            for val in enum:
                info_dict.update(
                    match_attribute_docs(
                        attr_name=val.name,
                        attr_docs=attr_docs,
                        attr_type_str=type(val.value).__name__,
                    )
                )
            handle_attributes_print(info_dict=info_dict, max_indent=max_indent)


def attrs_help(
    input_classes: Union[list, tuple],
    module_name: str,
    extract_fnc: Callable,
    max_indent: int,
) -> None:
    """Handles walking through a list classes to get help info

    For each class this function will search __doc__ and attempt to pull out help information for both the class
    itself and each attribute within the class. If it finds a repeated class in a iterable object it will
    recursively call self to handle information

    Args:
        input_classes: list of attr classes
        module_name: name of module to match
        extract_fnc: function that gets the nested lists within classes
        max_indent: max indent for pretty print of help

    Returns:
        None

    """
    # Handle the main loop
    other_list = handle_help_main(input_classes, module_name, extract_fnc, max_indent)
    handle_help_enums(
        other_list=other_list,
        module_name=module_name,
        extract_fnc=extract_fnc,
        max_indent=max_indent,
    )



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the base interface"""
import hashlib
import json
from abc import ABC, abstractmethod
from typing import Dict, Union

import attr

from spock.addons.tune.config import AxTunerConfig, OptunaTunerConfig
from spock.backend.wrappers import Spockspace


class BaseInterface(ABC):
    """Base interface for the various hyper-parameter tuner backends

    Attributes

        _tuner_config: spock version of the tuner configuration
        _tuner_namespace: tuner namespace that has attr classes that maps to an underlying library types

    """

    def __init__(self, tuner_config, tuner_namespace: Spockspace):
        """Base init call that maps a few variables

        Args:
            tuner_config: necessary dict object to determine the interface and sample correctly from the underlying library
            tuner_namespace: tuner namespace that has attr classes that maps to an underlying library types

        """
        self._tuner_config = tuner_config
        self._tuner_namespace = tuner_namespace

    @abstractmethod
    def sample(self):
        """Calls the underlying library sample to get a single sample/draw from the hyper-parameter
        sets (e.g. ranges, choices)

        Returns:
            Spockspace of the current hyper-parameter draw

        """

    @abstractmethod
    def _construct(self):
        """Constructs the base object needed by the underlying library to construct the correct object that allows
        for hyper-parameter sampling

        Returns:
            flat dictionary of all hyper-parameters named with dot notation (class.param_name)

        """

    @property
    @abstractmethod
    def _get_sample(self):
        """Gets the sample parameter dictionary from the underlying backend"""

    @property
    @abstractmethod
    def tuner_status(self):
        """Returns a dictionary of all the necessary underlying tuner internals to report the result"""

    @property
    @abstractmethod
    def best(self):
        """Returns a Spockspace of the best hyper-parameter config and the associated metric value"""

    @staticmethod
    def _sample_rollup(params):
        """Rollup the sample draw into a dictionary that can be converted to a spockspace with the correct names and
        roots -- un-dots the name structure

        Args:
            params: current parameter dictionary -- named by dot notation

        Returns:
            dictionary of rolled up sampled parameters
            md5 hash of the dictionary contents

        """
        key_set = {k.split(".")[0] for k in params.keys()}
        rollup_dict = {val: {} for val in key_set}
        for k, v in params.items():
            split_names = k.split(".")
            rollup_dict[split_names[0]].update({split_names[1]: v})
        dict_hash = hashlib.md5(
            json.dumps(rollup_dict, sort_keys=True).encode("utf-8")
        ).digest()
        return rollup_dict, dict_hash

    def _gen_spockspace(self, tune_dict: Dict):
        """Converts a dictionary of dictionaries of parameters into a valid Spockspace

        Args:
            tune_dict: dictionary of current parameters

        Returns:
            tune_dict: Spockspace

        """
        for k, v in tune_dict.items():
            attrs_dict = {
                ik: attr.ib(
                    validator=attr.validators.instance_of(type(iv)), type=type(iv)
                )
                for ik, iv in v.items()
            }
            obj = attr.make_class(name=k, attrs=attrs_dict, kw_only=True, frozen=True)
            tune_dict.update({k: obj(**v)})
        return self._to_spockspace(tune_dict)

    @staticmethod
    def _config_to_dict(tuner_config: Union[OptunaTunerConfig, AxTunerConfig]):
        """Turns an attrs config object into a dictionary

        Args:
            tuner_config: attrs config object

        Returns:
            dictionary of the attrs config object
        """
        return {k: v for k, v in attr.asdict(tuner_config).items() if v is not None}

    @staticmethod
    def _to_spockspace(tune_dict: Dict):
        """Converts a dict to a Spockspace

        Args:
            tune_dict: current dictionary

        Returns:
            Spockspace of dict

        """
        return Spockspace(**tune_dict)

    @staticmethod
    def _get_caster(val):
        """Gets a callable type object from a string type

        Args:
            val: current attr val:

        Returns:
            type class object

        """
        return __builtins__[val.type]

    def _try_choice_cast(self, val, type_string: str):
        """Try/except for casting choice parameters

        Args:
            val: current attr val
            type_string: spock hyper-parameter type name

        Returns:
            val: updated attr val

        """
        caster = self._get_caster(val)
        # Just attempt to cast in a try except
        try:
            val.choices = [caster(v) for v in val.choices]
            return val
        except Exception as e:
            raise TypeError(
                f"Attempted to cast into type: `{val.type}` but failed -- check the inputs to `{type_string}`: {e}"
            )

    def _try_range_cast(self, val, type_string: str):
        """Casting range parameters

        Note that we don't need to try/except here as the range is already constrained to be a float/int which
        will always be able to be cast into float/int

        Args:
            val: current attr val
            type_string: spock hyper-parameter type name

        Returns:
            low: low bound
            high: high bound

        """
        caster = self._get_caster(val)
        low = caster(val.bounds[0])
        high = caster(val.bounds[1])
        return low, high



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""
Spock is a framework that helps manage complex parameter configurations for Python applications

Please refer to the documentation provided in the README.md
"""

__all__ = [
    "builder",
    "config",
    "field_handlers",
    "handler",
    "help",
    "payload",
    "saver",
    "spaces",
    "typed",
    "utils",
    "wrappers",
]



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles prepping and saving the Spock config"""
from abc import abstractmethod
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
from uuid import uuid4

import attr

from spock.backend.handler import BaseHandler
from spock.backend.utils import _callable_2_str, _get_iter, _recurse_callables
from spock.backend.wrappers import Spockspace
from spock.utils import _T, add_info, get_packages


class BaseSaver(BaseHandler):  # pylint: disable=too-few-public-methods
    """Base class for saving configs

    Contains methods to build a correct output payload and then writes to file based on the file
    extension

    Attributes:
        _writers: maps file extension to the correct i/o handler
        _s3_config: optional S3Config object to handle s3 access

    """

    def __init__(self, s3_config: Optional[_T] = None):
        """Init function for base class

        Args:
            s3_config: optional s3Config object for S3 support
        """
        super(BaseSaver, self).__init__(s3_config=s3_config)

    def dict_payload(self, payload: Spockspace) -> Dict:
        """Clean up the config payload so that it can be returned as a dict representation

        Args:
            payload: dirty payload

        Returns:
            clean_dict: cleaned output payload

        """
        # Fix up values -- parameters
        return self._clean_up_values(payload)

    def save(
        self,
        payload: Spockspace,
        path: str,
        file_name: Optional[str] = None,
        create_save_path: bool = False,
        extra_info: bool = True,
        file_extension: str = ".yaml",
        tuner_payload: Optional[Spockspace] = None,
        fixed_uuid: Optional[str] = None,
    ) -> None:  # pylint: disable=too-many-arguments
        """Writes Spock config to file

        Cleans and builds an output payload and then correctly writes it to file based on the
        specified file extension

        Args:
            payload: current config payload
            path: path to save
            file_name: name of file (will be appended with .spock.cfg.file_extension) -- falls back to uuid if None
            create_save_path: boolean to create the path if non-existent
            extra_info: boolean to write extra info
            file_extension: what type of file to write
            tuner_payload: tuner level payload (unsampled)
            fixed_uuid: fixed uuid to allow for file overwrite

        Returns:
            None

        """
        # Check extension
        self._check_extension(file_extension=file_extension)
        # Make the filename -- always append uuid for unique-ness
        uuid_str = str(uuid4()) if fixed_uuid is None else fixed_uuid
        fname = "" if file_name is None else f"{file_name}."
        name = f"{fname}{uuid_str}.spock.cfg{file_extension}"
        # Fix up values -- parameters
        out_dict = self.dict_payload(payload)
        # Handle any env annotations that are present
        # Just stuff them into the dictionary
        crypto_flag = False
        for k, v in payload:
            if hasattr(v, "__resolver__"):
                for key, val in v.__resolver__.items():
                    out_dict[k][key] = val
            if hasattr(v, "__crypto__"):
                crypto_flag = True
        # Fix up the tuner values if present
        tuner_dict = (
            self._clean_tuner_values(tuner_payload)
            if tuner_payload is not None
            else None
        )
        if tuner_dict is not None:
            out_dict.update(tuner_dict)
        # Get extra info
        extra_dict = add_info() if extra_info else None
        library_dict = get_packages() if extra_info else None
        try:
            self._supported_extensions.get(file_extension)().save(
                out_dict=out_dict,
                info_dict=extra_dict,
                library_dict=library_dict,
                path=path,
                name=name,
                create_path=create_save_path,
                s3_config=self._s3_config,
                salt=payload.__salt__ if crypto_flag else None,
                key=payload.__key__ if crypto_flag else None,
            )
        except OSError as e:
            print(f"Unable to write to given path: {path / name}")
            raise e

    @abstractmethod
    def _clean_up_values(self, payload: Spockspace, remove_crypto: bool = True) -> Dict:
        """Clean up the config payload so it can be written to file

        Args:
            payload: dirty payload
            remove_crypto: try and remove crypto values if present

        Returns:
            clean_dict: cleaned output payload

        """

    @abstractmethod
    def _clean_tuner_values(self, payload: Spockspace) -> Dict:
        """Cleans up the base tuner payload that is not sampled

        Args:
            payload: dirty payload

        Returns:
            clean_dict: cleaned output payload

        """

    def _clean_output(self, out_dict: Dict) -> Dict:
        """Clean up the dictionary such that it can be written to file

        Args:
            out_dict: pre-cleaned dictionary

        Returns:
            clean_dict: cleaned output payload

        """
        clean_dict = {}
        for k, v in out_dict.items():
            if v is not None:
                clean_dict.update({k: self._recursive_tuple_2_list(v)})
        return clean_dict

    def _recursive_tuple_2_list(self, val: Any) -> Any:
        """Recursively find tuples and cast them to lists

        Args:
            val: current value of various types

        Returns:
            modified version of val

        """
        # If it is a tuple then we need to cast back -- do this first
        # so we can assign by idx unlike tuples
        if isinstance(val, (tuple, Tuple)):
            val = list(val)
        if isinstance(val, (list, List)):
            for idx, v in _get_iter(val):
                if isinstance(v, (dict, Dict, list, List, tuple, Tuple)):
                    val[idx] = self._recursive_tuple_2_list(v)
        elif isinstance(val, (dict, Dict)):
            new_dict = {}
            for k, v in _get_iter(val):
                if isinstance(v, (dict, Dict, list, List, tuple, Tuple)):
                    new_dict[k] = self._recursive_tuple_2_list(v)
                elif v is not None:
                    new_dict[k] = v
            val = new_dict
        return val


class AttrSaver(BaseSaver):
    """Base class for saving configs for the attrs backend

    Contains methods to build a correct output payload and then writes to file based on the file
    extension

    Attributes:
        _writers: maps file extension to the correct i/o handler

    """

    def __init__(self, s3_config: Optional[_T] = None):
        """Init for AttrSaver class

        Args:
            s3_config: s3Config object for S3 support
        """
        super().__init__(s3_config=s3_config)

    def __call__(self, *args, **kwargs):
        return AttrSaver(*args, **kwargs)

    def _clean_up_values(self, payload: Spockspace, remove_crypto: bool = True) -> Dict:
        # Dictionary to recursively write to
        out_dict = {}
        # All of the classes are defined at the top level
        all_spock_cls = set(vars(payload).keys())
        out_dict = self._recursively_handle_clean(
            payload, out_dict, all_cls=all_spock_cls
        )
        # Convert values
        clean_dict = self._clean_output(out_dict)
        # Clip any empty dictionaries
        clean_dict = {k: v for k, v in clean_dict.items() if len(v) > 0}
        if remove_crypto:
            if "__salt__" in clean_dict:
                _ = clean_dict.pop("__salt__")
            if "__key__" in clean_dict:
                _ = clean_dict.pop("__key__")
        return clean_dict

    def _clean_tuner_values(self, payload: Spockspace) -> Dict:
        # Just a double nested dict comprehension to unroll to dicts
        out_dict = {
            k: {ik: vars(iv) for ik, iv in vars(v).items()}
            for k, v in vars(payload).items()
        }
        # Convert values
        clean_dict = self._clean_output(out_dict)
        return clean_dict

    @staticmethod
    def _check_list_of_spock_classes(val: List, key: str, all_cls: Set) -> List:
        """Finds lists of spock classes and handles changing them to a writeable format

        Args:
            val: current value that is a list
            key: current dictionary key in the payload
            all_cls: set of all spock classes

        Returns:

        """
        # Check if each entry is a spock class
        clean_val = []
        repeat_flag = False
        for v in val:
            cls_name = type(v).__name__
            # For those that are a spock class and are repeated (cls_name == key) simply convert to dict
            if (cls_name in all_cls) and (cls_name == key):
                clean_val.append(attr.asdict(v))
            # For those whose cls is different than the key just append the cls name
            elif cls_name in all_cls:
                # Change the flag as this is a repeated class -- which needs to be compressed into a single
                # k:v pair
                repeat_flag = True
                clean_val.append(cls_name)
            # Fall back to the passed in values
            else:
                clean_val.append(v)
        # Handle repeated classes
        if repeat_flag:
            clean_val = list(set(clean_val))[-1]
        return clean_val

    def _recursively_handle_clean(
        self,
        payload: Spockspace,
        out_dict: Dict,
        parent_name: Optional[str] = None,
        all_cls: Optional[Set] = None,
    ) -> Dict:
        """Recursively works through spock classes and adds clean data to a dictionary

        Given a payload (Spockspace) work recursively through items that don't have parents to catch all
        parameter definitions while correctly mapping nested class definitions to their base level class thus
        allowing the output markdown to be a valid input file

        Args:
            payload: current payload (namespace)
            out_dict: output dictionary
            parent_name: name of the parent spock class if nested
            all_cls: all top level spock class definitions

        Returns:
            out_dict: modified dictionary with the cleaned data

        """
        for key, val in vars(payload).items():
            val_name = type(val).__name__
            # If v is any iterable we need to step into the nested structure
            if isinstance(val, (dict, Dict, list, List, tuple, Tuple)):
                # Need to inspect list type for lists of spock classes which must be handled differently
                mod_val = (
                    self._check_list_of_spock_classes(val, key, all_cls)
                    if isinstance(val, (list, List))
                    else val
                )
                # Recurses all iterables to find all callables and casts them to strings
                clean_val = _recurse_callables(
                    mod_val, _callable_2_str, check_type=Callable
                )
                out_dict.update({key: clean_val})
            # Catch any callables -- convert back to the str representation
            elif callable(val):
                out_dict.update({key: _callable_2_str(val)})
            # If it's a spock class but has a parent then just use the class name to reference the values
            elif (val_name in all_cls) and parent_name is not None:
                out_dict.update({key: val_name})
            # Check if it's a spock class without a parent -- iterate the values and recurse to catch more iterables
            elif val_name in all_cls:
                new_dict = self._recursively_handle_clean(
                    val, {}, parent_name=key, all_cls=all_cls
                )
                out_dict.update({key: new_dict})
            # Either base type or no nested values that could be Spock classes
            else:
                out_dict.update({key: val})
        return out_dict



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Resolver functions for Spock"""
import os
import re
from abc import ABC, abstractmethod
from distutils.util import strtobool
from typing import Any, ByteString, Optional, Pattern, Tuple, Union

from spock.backend.utils import decrypt_value
from spock.exceptions import _SpockResolverError
from spock.utils import _T


class BaseResolver(ABC):
    """Base class for resolvers

    Contains base methods for handling resolver syntax

    Attributes:
        _annotation_set: current set of supported resolver annotations

    """

    def __init__(self):
        """Init for BaseResolver class"""
        self._annotation_set = {"crypto", "inject"}

    @abstractmethod
    def resolve(self, value: Any, value_type: _T) -> Tuple[Any, Optional[str]]:
        """Resolves a variable from a given resolver syntax

        Args:
            value: current value to attempt to resolve
            value_type: type of the value to cast into

        Returns:
            Tuple of correctly typed resolved variable and any annotations

        """
        pass

    @staticmethod
    def _handle_default(value: str) -> Tuple[str, Union[str, None]]:
        """Handles setting defaults if allowed for a resolver

        Args:
            value: current string value

        Returns:
            tuple of given value and the default value

        """
        env_value, default_value = value.split(",")
        default_value = default_value.strip()
        # Swap string None to type None
        if default_value == "None":
            default_value = None
        return env_value, default_value

    @staticmethod
    def _check_base_regex(
        full_regex_op: Pattern,
        value: Any,
    ) -> bool:
        """Check if the value passed into the resolver matches the compiled regex op

        Args:
            full_regex_op: the full compiled regex
            value: the value passed into the resolver

        Returns:
            boolean if there is a regex match

        """
        # If it's a string we can check the regex
        if isinstance(value, str):
            # Check the regex and return non None status
            return full_regex_op.fullmatch(value) is not None
        # If it's not a string we can't resolve anything so just passthrough and let spock handle the value
        else:
            return False

    @staticmethod
    def _attempt_cast(maybe_env: Optional[str], value_type: _T, env_value: str) -> Any:
        """Attempts to cast the resolved variable into the given type

        Args:
            maybe_env: possible resolved variable
            value_type: type to cast into
            env_value: the reference to the resolved variable

        Returns:
            value type cast into the correct type

        Raises:
            _SpockResolverError if it cannot be cast into the specified type

        """
        # Attempt to cast in a try to be able to catch the failed type casts with an exception
        try:
            if value_type.__name__ == "bool":
                typed_env = (
                    value_type(strtobool(maybe_env)) if maybe_env is not None else False
                )
            else:
                typed_env = value_type(maybe_env) if maybe_env is not None else None
        except Exception as e:
            raise _SpockResolverError(
                f"Failed attempting to cast environment variable (name: {env_value}, value: `{maybe_env}`) "
                f"into Spock specified type `{value_type.__name__}`"
            )
        return typed_env

    def _apply_regex(
        self,
        end_regex_op: Pattern,
        clip_regex_op: Pattern,
        value: str,
        allow_default: bool,
        allow_annotation: bool,
    ) -> Tuple[str, str, Optional[str]]:
        """Applies the front and back regexes to the string value, determines defaults and annotations

        Args:
            end_regex_op: compiled regex for the back half of the match
            clip_regex_op: compiled regex for the front half of the match
            value: current string value to resolve
            allow_default: if allowed to contain default value syntax
            allow_annotation: if allowed to contain annotation syntax

        Returns:
            tuple containing the resolved string reference, the default value, and the annotation string

        Raises:
            _SpockResolverError if annotation isn't within the supported set, annotation is not supported, multiple `,`
                values are used, or defaults are given yet not supported

        """
        # Based on the start and end regex ops find the value the user set
        env_str = end_regex_op.split(clip_regex_op.split(value)[-1])[0]
        if (
            allow_annotation
            and len(clip_regex_op.split(value)) > 2
            and clip_regex_op.split(value)[1] != ""
        ):
            annotation = clip_regex_op.split(value)[1]
            if annotation not in self._annotation_set:
                raise _SpockResolverError(
                    f"Environment variable annotation must be within {self._annotation_set} -- got `{annotation}`"
                )
        elif (
            not allow_annotation
            and len(clip_regex_op.split(value)) > 2
            and clip_regex_op.split(value)[1] != ""
        ):
            raise _SpockResolverError(
                f"Found annotation style format however `{value}` does not support annotations"
            )
        else:
            annotation = None
        # Attempt to split on a comma for a default value
        split_len = len(env_str.split(","))
        # Default found if the len is 2
        if split_len == 2 and allow_default:
            env_value, default_value = self._handle_default(env_str)
        # If the length is larger than two then the syntax is messed up
        elif split_len > 2 and allow_default:
            raise _SpockResolverError(
                f"Issue with environment variable syntax -- currently `{value}` has more than one `,` which means the "
                f"optional default value cannot be resolved -- please use only one `,` separator within the syntax"
            )
        elif split_len > 1 and not allow_default:
            raise _SpockResolverError(
                f"Syntax does not support default values -- currently `{value}` contains the separator `,` which "
                f"id used to indicate default values"
            )
        else:
            env_value = env_str
            default_value = "None"
        return env_value, default_value, annotation


class EnvResolver(BaseResolver):
    """Class for resolving environmental variables

    Attributes:
        _annotation_set: current set of supported resolver annotations
        CLIP_ENV_PATTERN: regex for the front half
        CLIP_REGEX_OP: compiled regex for front half
        END_ENV_PATTERN: regex for back half
        END_REGEX_OP: comiled regex for back half
        FULL_ENV_PATTERN: full regex pattern
        FULL_REGEX_OP: compiled regex for full regex

    """

    # ENV Resolver -- full regex is ^\${spock\.env\.?([a-z]*?):.*}$
    CLIP_ENV_PATTERN = r"^\${spock\.env\.?([a-z]*?):"
    CLIP_REGEX_OP = re.compile(CLIP_ENV_PATTERN)
    END_ENV_PATTERN = r"}$"
    END_REGEX_OP = re.compile(END_ENV_PATTERN)
    FULL_ENV_PATTERN = CLIP_ENV_PATTERN + r".*" + END_ENV_PATTERN
    FULL_REGEX_OP = re.compile(FULL_ENV_PATTERN)

    def __init__(self):
        """Init for EnvResolver"""
        super(EnvResolver, self).__init__()

    def resolve(self, value: Any, value_type: _T) -> Tuple[Any, Optional[str]]:
        # Check the full regex for a match
        regex_match = self._check_base_regex(self.FULL_REGEX_OP, value)
        # if there is a regex match it needs to be handled by the underlying resolver ops
        if regex_match:
            # Apply the regex
            env_value, default_value, annotation = self._apply_regex(
                self.END_REGEX_OP,
                self.CLIP_REGEX_OP,
                value,
                allow_default=True,
                allow_annotation=True,
            )
            # Get the value from the env
            maybe_env = self._get_from_env(default_value, env_value)
            # Attempt to cast the value to its underlying type
            typed_env = self._attempt_cast(maybe_env, value_type, env_value)
        # Else just pass through
        else:
            typed_env = value
            annotation = None
        return typed_env, annotation

    @staticmethod
    def _get_from_env(default_value: Optional[str], env_value: str) -> Optional[str]:
        """Gets a value from an environmental variable

        Args:
            default_value: default value to fall back on for the env resolver
            env_value: current string of the env variable to get

        Returns:
            string or None for the resolved env variable

        Raises:
            _SpockResolverError if the env variable is not available or if no default is specified

        """
        # Attempt to get the env variable
        if default_value == "None":
            maybe_env = os.getenv(env_value)
        else:
            maybe_env = os.getenv(env_value, default_value)
        if maybe_env is None and default_value == "None":
            raise _SpockResolverError(
                f"Attempted to get `{env_value}` from environment variables but it is not set -- please set this "
                f"variable or provide a default via the following syntax ${{spock.env:{env_value},DEFAULT}}"
            )
        return maybe_env


class CryptoResolver(BaseResolver):
    """Class for resolving cryptographic variables

    Attributes:
        _annotation_set: current set of supported resolver annotations
        CLIP_ENV_PATTERN: regex for the front half
        CLIP_REGEX_OP: compiled regex for front half
        END_ENV_PATTERN: regex for back half
        END_REGEX_OP: comiled regex for back half
        FULL_ENV_PATTERN: full regex pattern
        FULL_REGEX_OP: compiled regex for full regex
        _salt: current cryptographic salt
        _key: current cryptographic key

    """

    # ENV Resolver -- full regex is ^\${spock\.crypto\.?([a-z]*?):.*}$
    CLIP_ENV_PATTERN = r"^\${spock\.crypto\.?([a-z]*?):"
    CLIP_REGEX_OP = re.compile(CLIP_ENV_PATTERN)
    END_ENV_PATTERN = r"}$"
    END_REGEX_OP = re.compile(END_ENV_PATTERN)
    FULL_ENV_PATTERN = CLIP_ENV_PATTERN + r".*" + END_ENV_PATTERN
    FULL_REGEX_OP = re.compile(FULL_ENV_PATTERN)

    def __init__(self, salt: str, key: ByteString):
        """Init for CryptoResolver

        Args:
            salt: cryptographic salt to use
            key: cryptographic key to use
        """
        super(CryptoResolver, self).__init__()
        self._salt = salt
        self._key = key

    def resolve(self, value: Any, value_type: _T) -> Tuple[Any, Optional[str]]:
        regex_match = self._check_base_regex(self.FULL_REGEX_OP, value)
        if regex_match:
            crypto_value, default_value, annotation = self._apply_regex(
                self.END_REGEX_OP,
                self.CLIP_REGEX_OP,
                value,
                allow_default=False,
                allow_annotation=False,
            )
            decrypted_value = decrypt_value(crypto_value, self._key, self._salt)
            typed_decrypted = self._attempt_cast(
                decrypted_value, value_type, crypto_value
            )
            annotation = "crypto"
        # Pass through
        else:
            typed_decrypted = value
            annotation = None
        # Crypto in --> crypto out annotation wise or else this exposes the value in plaintext
        return typed_decrypted, annotation



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""
Spock is a framework that helps manage complex parameter configurations for Python applications

Please refer to the documentation provided in the README.md
"""

from spock._version import get_versions
from spock.backend.typed import SavePath
from spock.builder import ConfigArgBuilder
from spock.config import spock

SpockBuilder = ConfigArgBuilder

__all__ = ["args", "builder", "config", "SavePath", "spock", "SpockBuilder"]

__version__ = get_versions()["version"]
del get_versions



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles registering field attributes for spock classes -- deals with the recursive nature of dependencies"""

import importlib
import sys
from abc import ABC, abstractmethod
from enum import EnumMeta
from typing import Any, ByteString, Callable, Dict, List, Tuple, Type

from attr import NOTHING, Attribute

from spock.backend.resolvers import CryptoResolver, EnvResolver
from spock.backend.spaces import AttributeSpace, BuilderSpace, ConfigSpace
from spock.backend.utils import (
    _get_name_py_version,
    _recurse_callables,
    _str_2_callable,
    encrypt_value,
)
from spock.exceptions import _SpockInstantiationError, _SpockNotOptionalError
from spock.utils import (
    _C,
    _T,
    _check_iterable,
    _is_spock_instance,
    _is_spock_tune_instance,
    _SpockVariadicGenericAlias,
)


class RegisterFieldTemplate(ABC):
    """Base class for handing different field types

    Once the configuration dictionary has been assembled from the config file and the command line then we need to
    map these values to the correct spock classes -- seeing as different types need to be handled differently and
    recursive calls might be needed (when referencing other spock classes) classes derived from RegisterFieldTemplate
    handle the logic for making sure the argument dictionary passes to the instantiation of each spock class is
    correct

    Attributes:
        special_keys: dictionary to check special keys

    """

    def __init__(self, salt: str, key: ByteString):
        """Init call for RegisterFieldTemplate class

        Args:
        """
        self.special_keys = {}
        self._salt = salt
        self._key = key
        self._env_resolver = EnvResolver()
        self._crypto_resolver = CryptoResolver(self._salt, self._key)

    def __call__(self, attr_space: AttributeSpace, builder_space: BuilderSpace):
        """Call method for RegisterFieldTemplate

        Handles calling the correct method for the type of the attribute

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        if self._is_attribute_in_config_arguments(attr_space, builder_space):
            self.handle_attribute_from_config(attr_space, builder_space)
        elif self._is_attribute_optional(attr_space.attribute):
            if isinstance(attr_space.attribute.default, type):
                self.handle_optional_attribute_type(attr_space, builder_space)
            else:
                self.handle_optional_attribute_value(attr_space, builder_space)

    def _is_attribute_in_config_arguments(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Checks if an attribute is in the configuration file or keyword arguments dictionary

        Will recurse spock classes as dependencies might be defined in the configs class

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: map of the read/cmd-line parameter dictionary to general or class level arguments

        Returns:
            boolean if in dictionary

        """
        # Instances might have other instances that might be defined in the configs
        # Recurse to try and catch all config defs
        # Only map if default is not None -- do so by evolving the attribute
        if (
            _is_spock_instance(attr_space.attribute.type)
            and attr_space.attribute.default is not None
        ):
            attr_space.field, special_keys = RegisterSpockCls(
                self._salt, self._key
            ).recurse_generate(
                attr_space.attribute.type, builder_space, self._salt, self._key
            )
            attr_space.attribute = attr_space.attribute.evolve(default=attr_space.field)
            builder_space.spock_space[
                attr_space.attribute.type.__name__
            ] = attr_space.field
            self.special_keys.update(special_keys)
        return (
            attr_space.config_space.name in builder_space.arguments
            and attr_space.attribute.name
            in builder_space.arguments[attr_space.config_space.name]
        )

    @staticmethod
    def _is_attribute_optional(attribute: Type[Attribute]):
        """Checks if an attribute is allowed to be optional

        Args:
            attribute: current attribute class

        Returns:
            boolean if the optional state is allowed

        """
        return (
            "optional" not in attribute.metadata and attribute.default is not NOTHING
        ) or ("optional" in attribute.metadata and attribute.metadata["optional"])

    def handle_optional_attribute_value(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles setting an optional value with its default

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """

        value, env_annotation = self._env_resolver.resolve(
            attr_space.attribute.default, attr_space.attribute.type
        )
        if env_annotation is not None:
            self._handle_env_annotations(
                attr_space, env_annotation, value, attr_space.attribute.default
            )
        value, crypto_annotation = self._crypto_resolver.resolve(
            value, attr_space.attribute.type
        )
        if crypto_annotation is not None:
            self._handle_crypto_annotations(
                attr_space, crypto_annotation, attr_space.attribute.default
            )
        attr_space.field = value

    def _handle_env_annotations(
        self, attr_space: AttributeSpace, annotation: str, value: Any, og_value: Any
    ):
        if annotation == "crypto":
            # Take the current value to string and then encrypt
            attr_space.annotations = (
                f"${{spock.crypto:{encrypt_value(str(value), self._key, self._salt)}}}"
            )
            attr_space.crypto = True
        elif annotation == "inject":
            attr_space.annotations = og_value
        else:
            raise _SpockInstantiationError(f"Got unknown env annotation `{annotation}`")

    @staticmethod
    def _handle_crypto_annotations(
        attr_space: AttributeSpace, annotation: str, og_value: str
    ):
        if annotation == "crypto":
            attr_space.annotations = og_value
            attr_space.crypto = True
        else:
            raise _SpockInstantiationError(
                f"Got unknown crypto annotation `{annotation}`"
            )

    @abstractmethod
    def handle_optional_attribute_type(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        pass

    @abstractmethod
    def handle_attribute_from_config(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        pass


class RegisterList(RegisterFieldTemplate):
    """Class that registers list types

    Attributes:
        special_keys: dictionary to check special keys

    """

    def __init__(self, salt: str, key: ByteString):
        """Init call to RegisterList

        Args:
        """
        super(RegisterList, self).__init__(salt, key)

    def handle_attribute_from_config(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles a list of spock config classes (aka repeated classes)

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        list_item_spock_class = attr_space.attribute.metadata["type"].__args__[0]
        attr_space.field = self._process_list(list_item_spock_class, builder_space)
        builder_space.spock_space[list_item_spock_class.__name__] = attr_space.field

    def handle_optional_attribute_type(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles a list of spock config classes (aka repeated classes) if it is optional

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        list_item_spock_class = attr_space.attribute.default
        attr_space.field = self._process_list(list_item_spock_class, builder_space)
        builder_space.spock_space[list_item_spock_class.__name__] = attr_space.field

    def handle_optional_attribute_value(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles setting the value for an optional basic attribute

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        super().handle_optional_attribute_value(attr_space, builder_space)
        if attr_space.field is not None:
            list_item_spock_class = attr_space.field
            # Here we need to catch the possibility of repeated lists via coded defaults
            if _is_spock_instance(attr_space.attribute.metadata["type"].__args__[0]):
                spock_cls = attr_space.attribute.metadata["type"].__args__[0]
                # Fall back to configs if present
                if spock_cls.__name__ in builder_space.arguments:
                    attr_space.field = self._process_list(spock_cls, builder_space)
                # Here we need to attempt to instantiate any class references that still exist
                try:
                    attr_space.field = [
                        val() if type(val) is type else val for val in attr_space.field
                    ]
                except Exception as e:
                    raise _SpockInstantiationError(
                        f"Spock class `{spock_cls.__name__}` could not be instantiated -- attrs message: {e}"
                    )
                builder_space.spock_space[spock_cls.__name__] = attr_space.field
            else:
                builder_space.spock_space[
                    list_item_spock_class.__name__
                ] = attr_space.field

    @staticmethod
    def _process_list(spock_cls, builder_space: BuilderSpace):
        """Rolls up repeated classes into the expected list format

        Args:
            spock_cls: current spock class
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
            list of rolled up repeated spock classes

        """
        return [
            spock_cls(**fields)
            for fields in builder_space.arguments[spock_cls.__name__]
        ]


class RegisterEnum(RegisterFieldTemplate):
    """Class that registers enum types

    Attributes:
        special_keys: dictionary to check special keys

    """

    def __init__(self, salt: str, key: ByteString):
        """Init call to RegisterEnum

        Args:
        """
        super(RegisterEnum, self).__init__(salt, key)

    def handle_attribute_from_config(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles getting the attribute set value when the Enum is made up of spock classes

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        possible_enum_classes = {
            c.value.__name__: c.value for c in attr_space.attribute.type
        }
        enum_cls_name = builder_space.arguments[attr_space.config_space.name][
            attr_space.attribute.name
        ]
        enum_cls = possible_enum_classes[enum_cls_name]
        self._handle_and_register_enum(enum_cls, attr_space, builder_space)

    def handle_optional_attribute_type(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles falling back on the optional default for a type based attribute

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        self._handle_and_register_enum(
            attr_space.attribute.default, attr_space, builder_space
        )

    def handle_optional_attribute_value(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles setting an optional value with its default

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        super().handle_optional_attribute_value(attr_space, builder_space)
        if attr_space.field is not None:
            builder_space.spock_space[
                type(attr_space.field).__name__
            ] = attr_space.field

    def _handle_and_register_enum(
        self, enum_cls, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Recurses the enum in case there are nested type definitions

        Args:
            enum_cls: current enum class
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        attr_space.field, special_keys = RegisterSpockCls.recurse_generate(
            enum_cls, builder_space, self._salt, self._key
        )
        self.special_keys.update(special_keys)
        builder_space.spock_space[enum_cls.__name__] = attr_space.field


class RegisterCallableField(RegisterFieldTemplate):
    """Class that registers callable types

    Attributes:
        special_keys: dictionary to check special keys

    """

    def __init__(self, salt: str, key: ByteString):
        """Init call to RegisterSimpleField

        Args:
        """
        super(RegisterCallableField, self).__init__(salt, key)

    def handle_attribute_from_config(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles setting a simple attribute when it is a spock class type

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        # These should always be strings
        str_field = builder_space.arguments[attr_space.config_space.name][
            attr_space.attribute.name
        ]
        call_ref = _str_2_callable(str_field)
        attr_space.field = call_ref

    def handle_optional_attribute_type(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Not implemented for this type

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Raises:
            _SpockNotOptionalError

        """
        raise _SpockNotOptionalError(
            f"Parameter `{attr_space.attribute.name}` within `{attr_space.config_space.name}` is of "
            f"type `{type(attr_space.attribute.type)}` which seems to be unsupported -- "
            f"are you missing an @spock decorator on a base python class?"
        )


class RegisterGenericAliasCallableField(RegisterFieldTemplate):
    """Class that registers Dicts containing callable types

    Attributes:
        special_keys: dictionary to check special keys

    """

    def __init__(self, salt: str, key: ByteString):
        """Init call to RegisterSimpleField

        Args:
        """
        super(RegisterGenericAliasCallableField, self).__init__(salt, key)

    def handle_attribute_from_config(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles setting a simple attribute when it is a spock class type

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        typed = attr_space.attribute.metadata["type"]
        out = None
        # Handle List Types
        if (
            _get_name_py_version(typed) == "List"
            or _get_name_py_version(typed) == "Tuple"
        ):
            out = []
            for v in builder_space.arguments[attr_space.config_space.name][
                attr_space.attribute.name
            ]:
                out.append(_recurse_callables(v, _str_2_callable))
        # Handle Dict Types
        elif _get_name_py_version(typed) == "Dict":
            out = {}
            for k, v in builder_space.arguments[attr_space.config_space.name][
                attr_space.attribute.name
            ].items():
                out.update({k: _recurse_callables(v, _str_2_callable)})
        attr_space.field = out

    def handle_optional_attribute_type(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Not implemented for this type

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Raises:
            _SpockNotOptionalError

        """
        raise _SpockNotOptionalError(
            f"Parameter `{attr_space.attribute.name}` within `{attr_space.config_space.name}` is of "
            f"type `{type(attr_space.attribute.type)}` which seems to be unsupported -- "
            f"are you missing an @spock decorator on a base python class?"
        )


class RegisterSimpleField(RegisterFieldTemplate):
    """Class that registers basic python types

    Attributes:
        special_keys: dictionary to check special keys

    """

    def __init__(self, salt: str, key: ByteString):
        """Init call to RegisterSimpleField

        Args:
        """
        super(RegisterSimpleField, self).__init__(salt, key)

    def handle_attribute_from_config(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles setting a simple attribute from a config file

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        og_value = builder_space.arguments[attr_space.config_space.name][
            attr_space.attribute.name
        ]
        value, env_annotation = self._env_resolver.resolve(
            og_value, attr_space.attribute.type
        )
        if env_annotation is not None:
            self._handle_env_annotations(attr_space, env_annotation, value, og_value)
        value, crypto_annotation = self._crypto_resolver.resolve(
            value, attr_space.attribute.type
        )
        if crypto_annotation is not None:
            self._handle_crypto_annotations(attr_space, crypto_annotation, og_value)
            attr_space.crypto = True
        attr_space.field = value
        self.register_special_key(attr_space)

    def handle_optional_attribute_type(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Not implemented for this type

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Raises:
            _SpockNotOptionalError

        """
        raise _SpockNotOptionalError(
            f"Parameter `{attr_space.attribute.name}` within `{attr_space.config_space.name}` is of "
            f"type `{type(attr_space.attribute.type)}` which seems to be unsupported -- "
            f"are you missing an @spock decorator on a base python class?"
        )

    def handle_optional_attribute_value(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles setting the attribute from default if optional

        Also checks for clashes with special keys

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        super().handle_optional_attribute_value(attr_space, builder_space)
        self.register_special_key(attr_space)

    def register_special_key(self, attr_space: AttributeSpace):
        """Registers a special key if it is found in the attribute metadata

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace

        Returns:
        """
        if (
            "special_key" in attr_space.attribute.metadata
            and attr_space.attribute.metadata["special_key"] is not None
        ):
            if attr_space.field is not None:
                self.special_keys["save_path"] = attr_space.field


class RegisterTuneCls(RegisterFieldTemplate):
    """Class that registers spock tune classes

    Attributes:
        special_keys: dictionary to check special keys

    """

    def __init__(self, salt: str, key: ByteString):
        """Init call to RegisterTuneCls

        Args:
        """
        super(RegisterTuneCls, self).__init__(salt, key)

    @staticmethod
    def _attr_type(attr_space: AttributeSpace):
        """Gets the attribute type

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace

        Returns:
            the type of the attribute

        """
        return attr_space.attribute.type

    def handle_attribute_from_config(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles when the spock tune class is made up of spock classes

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        attr_type = self._attr_type(attr_space)
        attr_space.field = attr_type(
            **builder_space.arguments[attr_space.config_space.name][
                attr_space.attribute.name
            ]
        )

    def handle_optional_attribute_value(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Not implemented for this type

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Raises:
            _SpockNotOptionalError

        """
        raise _SpockNotOptionalError()

    def handle_optional_attribute_type(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Not implemented for this type

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Raises:
            _SpockNotOptionalError

        """
        raise _SpockNotOptionalError()


class RegisterSpockCls(RegisterFieldTemplate):
    """Class that registers attributes within a spock class

    Might be called recursively so it has methods to deal with spock classes when invoked via the __call__ method

    Attributes:
        special_keys: dictionary to check special keys

    """

    def __init__(self, salt: str, key: ByteString):
        """Init call to RegisterSpockCls

        Args:
        """
        super(RegisterSpockCls, self).__init__(salt, key)

    @staticmethod
    def _attr_type(attr_space: AttributeSpace):
        """Gets the attribute type

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace

        Returns:
            the type of the attribute

        """
        return attr_space.attribute.type

    def handle_attribute_from_config(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles when the attribute is made up of a spock class or classes

        Calls the recurse_generate function which handles nesting of spock classes

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        attr_type = self._attr_type(attr_space)
        attr_space.field, special_keys = self.recurse_generate(
            attr_type, builder_space, self._salt, self._key
        )
        builder_space.spock_space[attr_type.__name__] = attr_space.field
        self.special_keys.update(special_keys)

    def handle_optional_attribute_value(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles when the falling back onto the default for the attribute of spock class type and the field value
        already exits within the attr_space

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        super().handle_optional_attribute_value(attr_space, builder_space)
        if attr_space.field is None:
            return
        builder_space.spock_space[
            self._attr_type(attr_space).__name__
        ] = attr_space.field

    def handle_optional_attribute_type(
        self, attr_space: AttributeSpace, builder_space: BuilderSpace
    ):
        """Handles when the falling back onto the default for the attribute of spock class type

        Calls the recurse_generate function which handles nesting of spock classes -- to make sure the attr_space.field
        value is defined

        Args:
            attr_space: holds information about a single attribute that is mapped to a ConfigSpace
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
        """
        attr_space.field, special_keys = RegisterSpockCls.recurse_generate(
            self._attr_type(attr_space), builder_space, self._salt, self._key
        )
        self.special_keys.update(special_keys)

        builder_space.spock_space[
            self._attr_type(attr_space).__name__
        ] = attr_space.field

    @classmethod
    def _find_callables(cls, typed: _T):
        """Attempts to find callables nested in Lists, Tuples, or Dicts

        Args:
            typed: input type

        Returns:
            boolean if callables are found

        """
        out = False
        if hasattr(typed, "__args__"):
            if (
                _get_name_py_version(typed) == "List"
                or _get_name_py_version(typed) == "Tuple"
            ):
                # Possibly nested Callables
                if not isinstance(typed.__args__[0], _SpockVariadicGenericAlias):
                    out = cls._find_callables(typed.__args__[0])
                # Found callables
                elif isinstance(typed.__args__[0], _SpockVariadicGenericAlias):
                    out = True
            elif _get_name_py_version(typed) == "Dict":
                key_type, value_type = typed.__args__
                if not isinstance(value_type, _SpockVariadicGenericAlias):
                    out = cls._find_callables(value_type)
                elif isinstance(value_type, _SpockVariadicGenericAlias):
                    out = True
            else:
                raise TypeError(
                    f"Unexpected type of `{str(typed)}` when attempting to handle GenericAlias types"
                )
        return out

    @classmethod
    def recurse_generate(
        cls, spock_cls: _C, builder_space: BuilderSpace, salt: str, key: ByteString
    ):
        """Call on a spock classes to iterate through the attrs attributes and handle each based on type and optionality

        Triggers a recursive call when an attribute refers to another spock classes

        Args:
            spock_cls: current spock class that is being handled
            builder_space: named_tuple containing the arguments and spock_space

        Returns:
            tuple of the instantiated spock class and the dictionary of special keys

        """
        # Empty dits for storing info
        special_keys = {}
        fields = {}
        annotations = {}
        crypto = False
        # Init the ConfigSpace for this spock class
        config_space = ConfigSpace(spock_cls, fields)
        # Iterate through the attrs within the spock class
        for attribute in spock_cls.__attrs_attrs__:
            attr_space = AttributeSpace(attribute, config_space)
            # Logic to handle the underlying type to call the correct Register* class
            # Lists of repeated values
            if (
                (attribute.type is list) or (attribute.type is List)
            ) and _is_spock_instance(attribute.metadata["type"].__args__[0]):
                handler = RegisterList(salt, key)
            # Dict/List of Callables
            elif (
                (attribute.type is list)
                or (attribute.type is List)
                or (attribute.type is dict)
                or (attribute.type is Dict)
                or (attribute.type is tuple)
                or (attribute.type is Tuple)
            ) and cls._find_callables(attribute.metadata["type"]):
                # handler = RegisterListCallableField()
                handler = RegisterGenericAliasCallableField(salt, key)
            # Enums
            elif isinstance(attribute.type, EnumMeta) and _check_iterable(
                attribute.type
            ):
                handler = RegisterEnum(salt, key)
            # References to other spock classes
            elif _is_spock_instance(attribute.type):
                handler = RegisterSpockCls(salt, key)
            # References to tuner classes
            elif _is_spock_tune_instance(attribute.type):
                handler = RegisterTuneCls(salt, key)
            # References to callables
            elif isinstance(attribute.type, _SpockVariadicGenericAlias):
                handler = RegisterCallableField(salt, key)
            # Basic field
            else:
                handler = RegisterSimpleField(salt, key)

            handler(attr_space, builder_space)
            special_keys.update(handler.special_keys)
            # Handle annotations by attaching them to a dictionary
            if attr_space.annotations is not None:
                annotations.update({attr_space.attribute.name: attr_space.annotations})
            if attr_space.crypto:
                crypto = True

        # Try except on the class since it might not be successful -- throw the attrs message as it will know the
        # error on instantiation
        try:
            # If there are annotations attach them to the spock class in the __resolver__ attribute
            if len(annotations) > 0:
                spock_cls.__resolver__ = annotations
            if crypto:
                spock_cls.__crypto__ = True
            spock_instance = spock_cls(**fields)
        except Exception as e:
            raise _SpockInstantiationError(
                f"Spock class `{spock_cls.__name__}` could not be instantiated -- attrs message: {e}"
            )
        return spock_instance, special_keys



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Utility functions for Spock"""

import ast
import os
import random
import socket
import subprocess
import sys
from argparse import _ArgumentGroup
from enum import EnumMeta
from math import isclose
from pathlib import Path
from time import localtime, strftime
from typing import Any, Dict, List, Tuple, Type, TypeVar, Union
from warnings import warn

import attr
import git
import pkg_resources

from spock.exceptions import _SpockValueError

minor = sys.version_info.minor


def make_salt(salt_len: int = 16):
    """Make a salt of specific length

    Args:
        salt_len: length of the constructed salt

    Returns:
        salt string

    """
    alphabet = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    return "".join(random.choice(alphabet) for _ in range(salt_len))


def _get_alias_type():
    """Gets the correct type of GenericAlias for versions less than 3.6

    Returns:
        _GenericAlias type

    """
    if minor < 7:
        from typing import GenericMeta as _GenericAlias
    else:
        from typing import _GenericAlias

    return _GenericAlias


def _get_callable_type():
    """Gets the correct underlying type reference for callable objects depending on the python version

    Returns:
        _VariadicGenericAlias type

    """
    if minor == 6:
        from typing import CallableMeta as _VariadicGenericAlias
    elif (minor > 6) and (minor < 9):
        from typing import _VariadicGenericAlias
    elif minor > 8:
        from typing import _CallableType as _VariadicGenericAlias
    else:
        raise RuntimeError(
            f"Attempting to use spock with python version `3.{minor}` which is unsupported"
        )
    return _VariadicGenericAlias


_SpockGenericAlias = _get_alias_type()
_SpockVariadicGenericAlias = _get_callable_type()
_T = TypeVar("_T")
_C = TypeVar("_C", bound=type)


def _filter_optional(val: List, allow_optional: bool = True):
    """Filters an iterable for None values if they are allowed

    Args:
        val: iterable of values that might contain None
        allow_optional: allows the check to succeed if a given val in the iterable is None

    Returns:
        filtered list of values with None values removed

    Raises:
        _SpockValueError

    """
    filtered_val = []
    for idx, v in enumerate(val):
        if v is None and not allow_optional:
            raise _SpockValueError(
                f"Optional values not allowed -- encountered None value in position `{idx}` of the iterable"
            )
        elif v is not None:
            filtered_val.append(v)
    return filtered_val


def sum_vals(
    val: List[Union[float, int, None]],
    sum_val: Union[float, int],
    allow_optional: bool = True,
    rel_tol: float = 1e-9,
    abs_tol: float = 0.0,
):
    """Checks if an iterable of values sums within tolerance to a specified value

    Args:
        val: iterable of values to sum
        sum_val: sum value to compare against
        allow_optional: allows the check to succeed if a given val in the iterable is None
        rel_tol: relative tolerance – it is the maximum allowed difference between a and b
        abs_tol: the minimum absolute tolerance – useful for comparisons near zero

    Returns:
        None

    Raises:
        _SpockValueError

    """
    filtered_val = _filter_optional(val, allow_optional)
    if not isclose(sum(filtered_val), sum_val, rel_tol=rel_tol, abs_tol=abs_tol):
        raise _SpockValueError(
            f"Sum of iterable is `{sum(filtered_val)}` which is not equal to specified value `{sum_val}` within given tolerances"
        )


def eq_len(val: List[Union[Tuple, List, None]], allow_optional: bool = True):
    """Checks that all values passed in the iterable are of the same length

    Args:
        val: iterable to compare lengths
        allow_optional: allows the check to succeed if a given val in the iterable is None

    Returns:
        None

    Raises:
        _SpockValueError

    """
    filtered_val = _filter_optional(val, allow_optional)
    # just do a set comprehension -- iterables shouldn't be that long so pay the O(n) price
    lens = {len(v) for v in filtered_val}
    if len(lens) != 1:
        raise _SpockValueError(
            f"Length mismatch -- current lengths of values within the iterable are `{lens}`"
        )


def within(
    val: Union[float, int, None],
    low_bound: Union[float, int],
    upper_bound: Union[float, int],
    inclusive_lower: bool = False,
    inclusive_upper: bool = False,
    allow_optional: bool = True,
) -> None:
    """Checks that a value is within a defined range

    Args:
        val: value to check against
        low_bound: lower bound of range
        upper_bound: upper bound of range
        inclusive_lower: if the check includes the bound value (i.e. >=)
        inclusive_upper: if the check includes the bound value (i.e. <=)
        allow_optional: allows the check to succeed if val is none

    Returns:
        None

    Raises:
        _SpockValueError

    """
    # Skip if None and allowed
    if allow_optional and val is None:
        pass
    elif val is None and not allow_optional:
        raise _SpockValueError(
            f"Set value is None and allow_optional is `{allow_optional}`"
        )
    else:
        # Check lower bounds
        lower_fn = le if inclusive_upper else lt
        upper_fn = ge if inclusive_lower else gt
        upper_fn(val=val, bound=low_bound, allow_optional=allow_optional)
        lower_fn(val=val, bound=upper_bound, allow_optional=allow_optional)


def ge(
    val: Union[float, int, None], bound: Union[float, int], allow_optional: bool = True
) -> None:
    """Checks that a value is greater than or equal to (inclusive) a lower bound

    Args:
        val: value to check against
        bound: lower bound
        allow_optional: allows the check to succeed if val is none

    Returns:
        None

    Raises:
        _SpockValueError

    """
    # Skip if None and allowed
    if allow_optional and val is None:
        pass
    elif val is None and not allow_optional:
        raise _SpockValueError(
            f"Set value is None and allow_optional is `{allow_optional}`"
        )
    else:
        if val < bound:
            raise _SpockValueError(
                f"Set value `{val}` is not >= given bound value `{bound}`"
            )


def gt(
    val: Union[float, int, None], bound: Union[float, int], allow_optional: bool = True
) -> None:
    """Checks that a value is greater (non-inclusive) than a lower bound

    Args:
        val: value to check against
        bound: lower bound
        allow_optional: allows the check to succeed if val is none

    Returns:
        None

    Raises:
        _SpockValueError

    """
    # Skip if None and allowed
    if allow_optional and val is None:
        pass
    elif val is None and not allow_optional:
        raise _SpockValueError(
            f"Set value is None and allow_optional is `{allow_optional}`"
        )
    else:
        if val <= bound:
            raise _SpockValueError(
                f"Set value `{val}` is not > given bound value `{bound}`"
            )


def le(
    val: Union[float, int, None], bound: Union[float, int], allow_optional: bool = True
) -> None:
    """Checks that a value is less than or equal to (inclusive) an upper bound

    Args:
        val: value to check against
        bound: upper bound
        allow_optional: allows the check to succeed if val is none

    Returns:
        None

    Raises:
        _SpockValueError

    """
    # Skip if None and allowed
    if allow_optional and val is None:
        pass
    elif val is None and not allow_optional:
        raise _SpockValueError(
            f"Set value is None and allow_optional is `{allow_optional}`"
        )
    else:
        if val > bound:
            raise _SpockValueError(
                f"Set value `{val}` is not <= given bound value `{bound}`"
            )


def lt(
    val: Union[float, int], bound: Union[float, int], allow_optional: bool = True
) -> None:
    """Checks that a value is less (non-inclusive) than an upper bound

    Args:
        val: value to check against
        bound: upper bound
        allow_optional: allows the check to succeed if val is none

    Returns:
        None

    Raises:
        _SpockValueError

    """
    # Skip if None and allowed
    if allow_optional and val is None:
        pass
    elif val is None and not allow_optional:
        raise _SpockValueError(
            f"Set value is None and allow_optional is `{allow_optional}`"
        )
    else:
        if val >= bound:
            raise _SpockValueError(
                f"Set value `{val}` is not < given bound value `{bound}`"
            )


def _find_all_spock_classes(attr_class: _C) -> List:
    """Within a spock class determine if there are any references to other spock classes

    Args:
        attr_class: a class with attrs attributes

    Returns:
        list of dependent spock classes

    """
    # Get the top level dict
    dict_attr = attr.fields_dict(attr_class)
    # Dependent classes
    dep_classes = []
    for k, v in dict_attr.items():
        # Checks for direct spock/attrs instance
        if _is_spock_instance(v.type):
            dep_classes.append(v.type)
        # Check for enum of spock/attrs instance
        elif isinstance(v.type, EnumMeta) and _check_4_spock_iterable(v.type):
            dep_classes.extend(_get_enum_classes(v.type))
        # Check for List[@spock-class] -- needs to be checked against 3.6 typing.List as well
        elif ((v.type is list) or (v.type is List)) and _is_spock_instance(
            v.metadata["type"].__args__[0]
        ):
            dep_classes.append(v.metadata["type"].__args__[0])
    return dep_classes


def _check_4_spock_iterable(iter_obj: Union[Tuple, List]) -> bool:
    """Checks if an iterable type contains a spock class

    Args:
        iter_obj: iterable type

    Returns:
        boolean if the iterable contains at least one spock class

    """
    return _check_iterable(iter_obj=iter_obj)


def _get_enum_classes(enum_obj: EnumMeta) -> List:
    """Checks if any of the values of an enum are spock classes and adds to a list

    Args:
        enum_obj: enum class

    Returns:
        list of enum values that are spock classes

    """
    return [v.value for v in enum_obj if _is_spock_instance(v.value)]


def path_object_to_s3path(path: Path) -> str:
    """Convert a path object into a string s3 path

    Args:
        path: a spock config path

    Returns:
        string of s3 path

    """
    return path.parts[0] + "//" + "/".join(path.parts[1:])


def check_path_s3(path: Path) -> bool:
    """Checks the given path to see if it matches the s3:// regex

    Args:
        path: a spock config path

    Returns:
        boolean of regex match

    """
    return len(path.parts) > 1 and path.parts[0] == "s3:"


def _is_spock_instance(__obj: object) -> bool:
    """Checks if the object is a @spock decorated class

    Private interface that checks to see if the object passed in is registered within the spock module and also
    is a class with attrs attributes (__attrs_attrs__)

    Args:
        __obj: class to inspect

    Returns:
        bool

    """
    return attr.has(__obj) and (__obj.__module__ == "spock.backend.config")


def _is_spock_tune_instance(__obj: object) -> bool:
    """Checks if the object is a @spock decorated class

    Private interface that checks to see if the object passed in is registered within the spock module tune addon and also
    is a class with attrs attributes (__attrs_attrs__)

    Args:
        __obj: class to inspect

    Returns:
        bool

    """
    return attr.has(__obj) and (__obj.__module__ == "spock.addons.tune.config")


def _check_iterable(iter_obj: Union[Tuple, List, EnumMeta]) -> bool:
    """Check if an iterable type contains a spock class

    Args:
        iter_obj: iterable type

    Returns:
        boolean if the iterable contains at least one spock class

    """
    return any([_is_spock_instance(v.value) for v in iter_obj])


def make_argument(
    arg_name: str, arg_type: _T, parser: Type[_ArgumentGroup]
) -> _ArgumentGroup:
    """Make argparser argument based on type

    Based on the type passed in handle the creation of the argparser argument so that overrides will have the correct
    behavior when set

    Args:
        arg_name: name for the argument
        arg_type: type of the argument
        parser: current parser

    Returns:
        parser: updated argparser

    """
    # For generic alias we take the input string and use a custom type callable to convert
    if isinstance(arg_type, _SpockGenericAlias):
        parser.add_argument(arg_name, required=False, type=_handle_generic_type_args)
    # For Unions -- python 3.6 can't deal with them correctly -- use the same ast method that generics require
    elif (
        hasattr(arg_type, "__origin__")
        and (arg_type.__origin__ is Union)
        and (minor < 7)
    ):
        parser.add_argument(arg_name, required=False, type=_handle_generic_type_args)
    # For choice enums we need to check a few things first
    elif isinstance(arg_type, EnumMeta):
        type_set = list({type(val.value) for val in arg_type})[0]
        # if this is an enum of a class switch the type to str as this is how it gets matched
        type_set = str if type_set.__name__ == "type" else type_set
        parser.add_argument(arg_name, required=False, type=type_set)
    # For booleans we map to store true
    elif arg_type == bool:
        parser.add_argument(arg_name, required=False, action="store_true")
    # Else we are a simple base type which we can cast to
    else:
        parser.add_argument(arg_name, required=False, type=arg_type)
    return parser


def _handle_generic_type_args(val: str) -> Any:
    """Evaluates a string containing a Python literal

    Seeing a list and tuple types will come in as string literal format, use ast to get the actual type

    Args:
        val: string literal

    Returns:
        the underlying string literal type

    """
    return ast.literal_eval(val)


def get_packages() -> Dict:
    """Gets all currently installed packages and assembles a dictionary of name: version

    Notes:
        https://stackoverflow.com/a/50013400

    Returns:
        dictionary of all currently available packages

    """
    named_list = sorted([str(i.key) for i in pkg_resources.working_set])
    return {
        f"# {i}": str(pkg_resources.working_set.by_key[i].version) for i in named_list
    }


def add_info() -> Dict:
    """Adds extra information to the output dictionary

    Args:

    Returns:
        out_dict: output dictionary
    """
    out_dict = {}
    out_dict = add_generic_info(out_dict)
    out_dict = add_repo_info(out_dict)
    return out_dict


def make_blank_git(out_dict: Dict) -> Dict:
    """Adds blank git info

    Args:
        out_dict: current output dictionary

    Returns:
        out_dict: output dictionary with added git info

    """
    for key in ("BRANCH", "COMMIT SHA", "STATUS", "ORIGIN"):
        out_dict.update({f"# Git {key}": "UNKNOWN"})
    return out_dict


def add_repo_info(out_dict: Dict) -> Dict:
    """Adds GIT information to the output dictionary

    Args:
        out_dict: output dictionary

    Returns:
        out_dict: output dictionary
    """
    try:  # pragma: no cover
        # Assume we are working out of a repo
        repo = git.Repo(os.getcwd(), search_parent_directories=True)
        # Check if we are really in a detached head state as later info will fail if we are
        if minor < 7:
            head_result = subprocess.run(
                "git rev-parse --abbrev-ref --symbolic-full-name HEAD",
                stdout=subprocess.PIPE,
                shell=True,
                check=False,
            )
        else:
            head_result = subprocess.run(
                "git rev-parse --abbrev-ref --symbolic-full-name HEAD",
                capture_output=True,
                shell=True,
                check=False,
            )
        if head_result.stdout.decode().rstrip("\n") == "HEAD":
            out_dict = make_blank_git(out_dict)
        else:
            out_dict.update({"# Git Branch": repo.active_branch.name})
            out_dict.update({"# Git Commit": repo.active_branch.commit.hexsha})
            out_dict.update(
                {"# Git Date": repo.active_branch.commit.committed_datetime}
            )
            if (
                len(repo.untracked_files) > 0
                or len(repo.active_branch.commit.diff(None)) > 0
            ):
                git_status = "DIRTY"
            else:
                git_status = "CLEAN"
            out_dict.update({"# Git Status": git_status})
            out_dict.update(
                {"# Git Origin": repo.active_branch.commit.repo.remotes.origin.url}
            )
    except git.InvalidGitRepositoryError:  # pragma: no cover
        # But it's okay if we are not
        out_dict = make_blank_git(out_dict)
    return out_dict


def add_generic_info(out_dict: Dict) -> Dict:
    """Adds date, fqdn information to the output dictionary

    Args:
        out_dict: output dictionary

    Returns:
        out_dict: output dictionary
    """
    out_dict.update({"# Machine FQDN": socket.getfqdn()})
    out_dict.update({"# Python Executable": sys.executable})
    out_dict.update(
        {
            "# Python Version": f"{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}"
        }
    )
    out_dict.update({"# Python Script": os.path.realpath(sys.argv[0])})
    out_dict.update({"# Run Date": strftime("%Y-%m-%d", localtime())})
    out_dict.update({"# Run Time": strftime("%H:%M:%S", localtime())})
    # Make a best effort to determine if run in a container
    out_dict.update({"# Run w/ Docker": str(_maybe_docker())})
    # Make a best effort to determine if run in a container via k8s
    out_dict.update({"# Run w/ Kubernetes": str(_maybe_k8s())})

    return out_dict


def _maybe_docker(cgroup_path: str = "/proc/self/cgroup") -> bool:
    """Make a best effort to determine if run in a docker container

    Args:
        cgroup_path: path to cgroup file

    Returns:
        boolean of best effort docker determination

    """
    # A few options seem to be at play here:
    # 1. Check for /.dockerenv -- docker should create this is any container
    bool_env = os.path.exists("/.dockerenv")
    # 2. Check /proc/self/cgroup for "docker"
    # https://stackoverflow.com/a/48710609
    bool_cgroup = os.path.isfile(cgroup_path) and any(
        "docker" in line for line in open(cgroup_path)
    )
    return bool_env or bool_cgroup


def _maybe_k8s(cgroup_path: str = "/proc/self/cgroup") -> bool:
    """Make a best effort to determine if run in a container via k8s

    Args:
        cgroup_path: path to cgroup file

    Returns:
        boolean of best effort k8s determination

    """
    # A few options seem to be at play here:
    # 1. Check for KUBERNETES_SERVICE_HOST -- kublet should add this to every running pod
    bool_env = os.environ.get("KUBERNETES_SERVICE_HOST") is not None
    # 2. Similar to docker check /proc/self/cgroup for "kubepods"
    # https://stackoverflow.com/a/48710609
    bool_cgroup = os.path.isfile(cgroup_path) and any(
        "kubepods" in line for line in open(cgroup_path)
    )
    return bool_env or bool_cgroup


def deep_payload_update(source: Dict, updates: Dict) -> Dict:
    """Deeply updates a dictionary

    Iterates through a dictionary recursively to update individual values within a possibly nested dictionary
    of dictionaries -- creates a dictionary if empty and trying to recurse

    Args:
        source: source dictionary
        updates: updates to the dictionary

    Returns:
        source: updated version of the source dictionary

    """

    for k, v in updates.items():
        if isinstance(v, (dict, Dict)) and v:
            source_dict = {} if source.get(k) is None else source.get(k)
            updated_dict = deep_payload_update(source_dict, v)
            if updated_dict:
                source[k] = updated_dict
        else:
            source[k] = v
    return source


def check_payload_overwrite(
    payload: Dict, updates: Dict, configs: str, overwrite: str = ""
) -> None:
    """Warns when parameters are overwritten across payloads as order will matter

    Args:
        payload: current payload
        payload_update: update to add to payload
        configs: config path
        overwrite: name of parent

    Returns:
    """
    for k, v in updates.items():
        if isinstance(v, dict) and v:
            overwrite += k + ":"
            current_payload = {} if payload.get(k) is None else payload.get(k)
            check_payload_overwrite(current_payload, v, configs, overwrite=overwrite)
        else:
            if k in payload:
                warn(
                    f"Overriding an already set parameter {overwrite + k} from {configs}\n"
                    f"Be aware that value precedence is set by the order of the config files (last to load)...",
                    SyntaxWarning,
                )



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""
Spock is a framework that helps manage complex parameter configurations for Python applications

Please refer to the documentation provided in the README.md
"""


__all__ = ["s3", "tune"]



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles creation and ops for DAGs"""

import sys
from typing import Dict, Generator, List, Tuple

from spock.utils import _find_all_spock_classes


class Graph:
    """Class that holds graph methods for determining dependencies between spock classes

    Attributes:
        _input_classes: list of input classes that link to a backend
        _dag: graph of the dependencies between spock classes
        _lazy: attempts to lazily find @spock decorated classes registered within sys.modules["spock"].backend.config

    """

    def __init__(self, input_classes: List, lazy: bool):
        """Init call for Graph class

        Args:
            input_classes: list of input classes that link to a backend
            lazy: attempts to lazily find @spock decorated classes registered within sys.modules["spock"].backend.config
        """
        self._input_classes = input_classes
        self._lazy = lazy
        # Maybe find classes lazily -- roll them into the input class tuple
        # make sure to cast as a set first since the lazy search might find duplicate references
        if self._lazy:
            # Lazily find base classes
            self._input_classes = (
                *self._input_classes,
                *set(self._lazily_find_classes(self._input_classes)),
            )
            # Lazily find any parents that are missing
            self._input_classes = (
                *self._input_classes,
                *set(self._lazily_find_parents()),
            )
        # Build -- post lazy eval
        self._dag = self._build()
        # Validate (No cycles in DAG)
        if self._has_cycles() is True:
            raise ValueError(
                "Cycle detected within the spock class dependency DAG...\n"
                "Please correct your @spock decorated classes by removing any cyclic references"
            )

    @property
    def dag(self):
        """Returns the DAG"""
        return self._dag

    @property
    def nodes(self):
        """Returns the input_classes/nodes"""
        return self._input_classes

    @property
    def node_names(self):
        """Returns the node names"""
        return {f"{k.__name__}" for k in self.nodes}

    @property
    def node_map(self):
        return {f"{k.__name__}": k for k in self.nodes}

    @property
    def roots(self):
        """Returns the roots of the dependency graph"""
        return [self.node_map[k] for k, v in self.dag.items() if len(v) == 0]

    @property
    def topological_order(self):
        return self._topological_sort()

    @staticmethod
    def _yield_class_deps(classes: List) -> Generator[Tuple, None, None]:
        """Generator to iterate through nodes and find dependencies

        Args:
            classes: list of classes to iterate through

        Yields:
            tuple or the base input class and the current name of the dependent class

        """
        for input_class in classes:
            dep_names = {f"{v.__name__}" for v in _find_all_spock_classes(input_class)}
            for v in dep_names:
                yield input_class, v

    def _lazily_find_classes(self, classes: List) -> Tuple:
        """Searches within the spock sys modules attributes to lazily find @spock decorated classes

        These classes have been decorated with @spock but might not have been passes into the ConfigArgBuilder so
        this allows for 'lazy' lookup of these classes to make the call to ConfigArgBuilder a little less verbose
        when there are a lot of spock classes

        Returns:
            tuple of any lazily discovered classes

        """
        # Iterate thorough all of the base spock classes to get the dependencies and reverse dependencies
        lazy_classes = []
        for _, v in self._yield_class_deps(classes):
            if (
                hasattr(sys.modules["spock"].backend.config, v)
                and getattr(sys.modules["spock"].backend.config, v) not in classes
            ):
                print(
                    f"Lazy evaluation found a @spock decorated class named `{v}` within the registered types of "
                    f"sys.modules['spock'].backend.config -- Attempting to use the class "
                    f"`{getattr(sys.modules['spock'].backend.config, v)}` within the SpockBuilder"
                )
                # Get the lazily discovered class
                lazy_class = getattr(sys.modules["spock"].backend.config, v)
                # Recursive check the lazy class for other lazy classes
                dependent_lazy_classes = self._lazily_find_classes([lazy_class])
                # extend the list if the recursive check finds any other lazy classes
                if len(dependent_lazy_classes) > 0:
                    lazy_classes.extend(dependent_lazy_classes)
                lazy_classes.append(lazy_class)
        return tuple(lazy_classes)

    def _lazily_find_parents(self) -> Tuple:
        """Searches within the current set of input_classes (@spock decorated classes) to lazily find any parents

        Given that lazy inheritance means that the parent classes won't be included (since they are cast to spock
        classes within the decorator and the MRO is handled internally) this allows the lazy flag to find those parent
        classes and add them to the SpockBuilder *args (input classes).

        Returns:
            tuple of any lazily discovered classes

        """
        lazy_parents = {}
        for v in self._input_classes:
            # Check if the MRO is long enough to have bases
            if len(v.__mro__[1:-1]) > 0:
                bases = list(v.__mro__[1:-1])
                for base in bases:
                    cls_name = base.__name__
                    if (cls_name not in self.node_names) and (
                        cls_name not in lazy_parents.keys()
                    ):
                        print(
                            f"Lazy evaluation found a @spock decorated parent class named `{cls_name}` within the "
                            f"registered types of sys.modules['spock'].backend.config -- Appending the class "
                            f"`{getattr(sys.modules['spock'].backend.config, cls_name)}` to the SpockBuilder..."
                        )
                        lazy_parents.update({base.__name__: base})
        return tuple(lazy_parents.values())

    def _build(self) -> Dict:
        """Builds a dictionary of nodes and their edges (essentially builds the DAG)

        Returns:
            dictionary of nodes and their edges

        """
        # Build a dictionary of all nodes (base spock classes)
        nodes = {val: [] for val in self.node_names}
        # Iterate thorough all of the base spock classes to get the dependencies and reverse dependencies
        for input_class, v in self._yield_class_deps(self._input_classes):
            if v not in self.node_names:
                raise ValueError(
                    f"Missing @spock decorated class -- `{v}` was not passed as an *arg to "
                    f"ConfigArgBuilder and/or could not be found via lazy evaluation (currently lazy=`{self._lazy}`) "
                    f"within sys.modules['spock'].backend.config"
                )
            nodes.get(v).append(input_class)
        nodes = {key: set(val) for key, val in nodes.items()}
        return nodes

    def _has_cycles(self) -> bool:
        """Uses DFS to check for cycles within the spock dependency graph

        Returns:
            boolean if a cycle is found

        """
        # DFS w/ recursion stack for DAG cycle detection
        visited = {key: False for key in self.dag.keys()}
        all_nodes = list(visited.keys())
        recursion_stack = {key: False for key in self.dag.keys()}
        # Recur for all edges
        for node in all_nodes:
            if visited.get(node) is False:
                # Surface the recursive checks
                if self._cycle_dfs(node, visited, recursion_stack) is True:
                    return True
        return False

    def _cycle_dfs(self, node: str, visited: Dict, recursion_stack: Dict) -> bool:
        """DFS via a recursion stack for cycles

        Args:
            node: current graph node (spock class type)
            visited: dictionary of visited nodes
            recursion_stack: dictionary that is the recursion stack that is used to find cycles

        Returns:
            boolean if a cycle is found

        """
        # Update the visited nodes
        visited.update({node: True})
        # Update recursion stack
        recursion_stack.update({node: True})
        # Recur through the edges
        for val in self.dag.get(node):
            if visited.get(val) is False:
                # The the recursion returns True then work it up the stack
                if self._cycle_dfs(val, visited, recursion_stack) is True:
                    return True
            # If the vertex is already in the recursion stack then we have a cycle
            elif recursion_stack.get(val) is True:
                return True
        # Reset the stack for the current node if we've completed the DFS from this node
        recursion_stack.update({node: False})
        return False

    def _topological_sort(self) -> List:
        # DFS for topological sort
        # https://en.wikipedia.org/wiki/Topological_sorting
        visited = {key: False for key in self.node_names}
        all_nodes = list(visited.keys())
        stack = []
        for node in all_nodes:
            if visited.get(node) is False:
                self._topological_sort_dfs(node, visited, stack)
        stack.reverse()
        return stack

    def _topological_sort_dfs(self, node: str, visited: Dict, stack: List) -> None:
        # Update the visited dict
        visited.update({node: True})
        # Recur for all edges
        for val in self._dag.get(node):
            if visited.get(val.__name__) is False:
                self._topological_sort_dfs(val.__name__, visited, stack)
        stack.append(node)



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Creates the spock config interface that wraps attr"""

import sys

import attr

from spock.backend.typed import katra
from spock.exceptions import _SpockUndecoratedClass
from spock.utils import _is_spock_instance


def _base_attr(cls, kw_only, make_init, dynamic):
    """Map type hints to katras

    Connector function that maps type hinting style to the defined katra style which uses the more strict
    attr.ib() definition

    Handles dynamic decorators which allows for inheritance of non @spock decorated classes

    Args:
        cls: basic class def
        dynamic: allows inherited classes to not be @spock decorated

    Returns:
        cls: base spock classes derived from the MRO
        attrs_dict: the current dictionary of attr.attribute values
        merged_annotations: dictionary of type annotations

    """
    # Since we are not using the @attr.s decorator we need to get the parent classes for inheritance
    # We do this by using the mro and grabbing anything that is not the first and last indices in the list and wrapping
    # it into a tuple
    bases = ()
    base_annotation = {}
    base_defaults = {}
    if len(cls.__mro__[1:-1]) > 0:
        # Get bases minus self and python class object
        bases = list(cls.__mro__[1:-1])
        for idx, base_cls in enumerate(bases):
            if not _is_spock_instance(base_cls) and not dynamic:
                raise _SpockUndecoratedClass(
                    f"Class `{base_cls.__name__}` was not decorated with the @spock decorator "
                    f"and `dynamic={dynamic}` was set for child class `{cls.__name__}` -- Please remedy one of these"
                )
            elif not _is_spock_instance(base_cls) and dynamic:
                bases[idx] = _process_class(base_cls, kw_only, make_init, dynamic)
        bases = tuple(bases)
        base_annotation = {}
        for val in bases:
            for attribute in val.__attrs_attrs__:
                # Since we are moving left to right via the MRO only update if not currently present
                # this maintains parity to how the MRO is handled in base python
                if attribute.name not in base_annotation:
                    if "type" in attribute.metadata:
                        base_annotation.update(
                            {attribute.name: attribute.metadata["og_type"]}
                        )
                    else:
                        base_annotation.update(
                            {attribute.name: val.__annotations__[attribute.name]}
                        )
        base_defaults = {
            attribute.name: attribute.default
            for val in bases
            for attribute in val.__attrs_attrs__
            if attribute.default is not (None or attr.NOTHING)
        }
    # Merge the annotations -- always override as this is the lowest level of the MRO
    if hasattr(cls, "__annotations__"):
        new_annotations = {k: v for k, v in cls.__annotations__.items()}
    else:
        new_annotations = {}
    merged_annotations = {**base_annotation, **new_annotations}

    # Make a blank attrs dict for new attrs
    attrs_dict = {}

    for k, v in merged_annotations.items():
        # If the cls has the attribute then a default was set
        if hasattr(cls, k):
            default = getattr(cls, k)
        elif k in base_defaults:
            default = base_defaults[k]
        else:
            default = None
        attrs_dict.update({k: katra(typed=v, default=default)})
    return bases, attrs_dict, merged_annotations


def _process_class(cls, kw_only: bool, make_init: bool, dynamic: bool):
    """Process a given class

    Args:
        cls: basic class definition
        kw_only: set kwarg only
        make_init: make an init function
        dynamic: allows inherited classes to not be @spock decorated

    Returns:
        cls with attrs dunder methods added

    """
    # Handles the MRO and gets old annotations
    bases, attrs_dict, merged_annotations = _base_attr(cls, kw_only, make_init, dynamic)
    # Copy over the post init function -- borrow a bit from attrs library to add the __post__hook__ method to the
    # init call via `"__attrs_post_init__"`
    if hasattr(cls, "__post_hook__"):
        attrs_dict.update({"__attrs_post_init__": cls.__post_hook__})
    # Dynamically make an attr class
    obj = attr.make_class(
        name=cls.__name__,
        bases=bases,
        attrs=attrs_dict,
        kw_only=kw_only,
        frozen=True,
        auto_attribs=True,
        init=make_init,
    )
    # For each class we dynamically create we need to register it within the system modules for pickle to work
    setattr(sys.modules["spock"].backend.config, obj.__name__, obj)
    # Swap the __doc__ string from cls to obj
    obj.__doc__ = cls.__doc__
    # Set the __init__ function
    # Handle __annotations__ from the MRO
    obj.__annotations__ = merged_annotations
    return obj


def spock_attr(
    maybe_cls=None,
    kw_only=True,
    make_init=True,
    dynamic=False,
):
    """Map type hints to katras

    Connector function that maps type hinting style to the defined katra style which uses the more strict
    attr.ib() definition -- this allows us to attach the correct validators for types before the attrs class is
    built

    Args:
        maybe_cls: maybe a basic class def maybe None depending on call type
        kw_only: Make all attributes keyword-only
        make_init: bool, define a __init__() method
        dynamic: allows inherited classes to not be @spock decorated -- will automatically cast parent classes to a
            spock class by traversing the MRO

    Returns:
        cls: attrs class that is frozen and kw only
    """

    def wrap(cls):
        return _process_class(
            cls, kw_only=kw_only, make_init=make_init, dynamic=dynamic
        )

    # Note: Taken from dataclass/attr definition(s)
    # maybe_cls's type depends on the usage of the decorator.  It's a class
    # if it's used as `@spock` but ``None`` if used as `@spock()`.
    if maybe_cls is None:
        return wrap
    else:
        return wrap(maybe_cls)



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the ax backend"""

from ax.service.ax_client import AxClient

from spock.addons.tune.config import AxTunerConfig
from spock.addons.tune.interface import BaseInterface

try:
    from typing import TypedDict
except ImportError:
    from mypy_extensions import TypedDict


class AxTunerStatus(TypedDict):
    """Tuner status return object for Ax -- supports the service style API from Ax

    Attributes:
        client: current AxClient instance
        trial_index: current trial index

    """

    client: AxClient
    trial_index: int


class AxInterface(BaseInterface):
    """Specific override to support the Ax backend -- supports the service style API from Ax"""

    def __init__(self, tuner_config: AxTunerConfig, tuner_namespace):
        """AxInterface init call that maps variables, creates a map to fnc calls, and constructs the necessary
        underlying objects

        Args:
            tuner_config: configuration object for the ax backend
            tuner_namespace: tuner namespace that has attr classes that maps to an underlying library types
        """
        super(AxInterface, self).__init__(tuner_config, tuner_namespace)
        self._tuner_obj = AxClient(
            generation_strategy=self._tuner_config.generation_strategy,
            enforce_sequential_optimization=self._tuner_config.enforce_sequential_optimization,
            random_seed=self._tuner_config.random_seed,
            verbose_logging=self._tuner_config.verbose_logging,
        )
        # Some variables to use later
        self._trial_index = None
        self._sample_hash = None
        # Mapping spock underlying classes to ax distributions (search space)
        self._map_type = {
            "RangeHyperParameter": {
                "int": self._ax_range,
                "float": self._ax_range,
            },
            "ChoiceHyperParameter": {
                "int": self._ax_choice,
                "float": self._ax_choice,
                "str": self._ax_choice,
                "bool": self._ax_choice,
            },
        }
        # Build the correct underlying dictionary object for Ax client create experiment
        self._param_obj = self._construct()
        # Create the AxClient experiment
        self._tuner_obj.create_experiment(
            parameters=self._param_obj,
            name=self._tuner_config.name,
            objective_name=self._tuner_config.objective_name,
            minimize=self._tuner_config.minimize,
            parameter_constraints=self._tuner_config.parameter_constraints,
            outcome_constraints=self._tuner_config.outcome_constraints,
            overwrite_existing_experiment=self._tuner_config.overwrite_existing_experiment,
            tracking_metric_names=self._tuner_config.tracking_metric_names,
            immutable_search_space_and_opt_config=self._tuner_config.immutable_search_space_and_opt_config,
            is_test=self._tuner_config.is_test,
        )

    @property
    def tuner_status(self) -> AxTunerStatus:
        return AxTunerStatus(client=self._tuner_obj, trial_index=self._trial_index)

    @property
    def best(self):
        best_obj = self._tuner_obj.get_best_parameters()
        rollup_dict, _ = self._sample_rollup(best_obj[0])
        return (
            self._gen_spockspace(rollup_dict),
            best_obj[1][0][self._tuner_obj.objective_name],
        )

    @property
    def _get_sample(self):
        return self._tuner_obj.get_next_trial()

    def sample(self):
        parameters, self._trial_index = self._get_sample
        # Roll this back out into a Spockspace so it can be merged into the fixed parameter Spockspace
        # Also need to un-dot the param names to rebuild the nested structure
        rollup_dict, sample_hash = self._sample_rollup(parameters)
        self._sample_hash = sample_hash
        return self._gen_spockspace(rollup_dict)

    def _construct(self):
        param_list = []
        # These will only be nested one level deep given the tuner syntax
        for k, v in vars(self._tuner_namespace).items():
            for ik, iv in vars(v).items():
                param_fn = self._map_type[type(iv).__name__][iv.type]
                param_list.append(param_fn(name=f"{k}.{ik}", val=iv))
        return param_list

    def _ax_range(self, name, val):
        """Assemble the dictionary for ax range parameters

        Args:
            name: parameter name
            val: current attr val

        Returns:
            dictionary that can be added to a parameter list

        """
        low, high = self._try_range_cast(val, type_string="RangeHyperParameter")
        return {
            "name": name,
            "type": "range",
            "bounds": [low, high],
            "value_type": val.type,
            "log_scale": val.log_scale,
        }

    def _ax_choice(self, name, val):
        """Assemble the dictionary for ax choice parameters

        Args:
            name: parameter name
            val: current attr val

        Returns:
            dictionary that can be added to a parameter list

        """
        val = self._try_choice_cast(val, type_string="ChoiceHyperParameter")
        return {
            "name": name,
            "type": "choice",
            "values": val.choices,
            "value_type": val.type,
        }



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Handles the building/saving of the configurations from the Spock config classes"""

import argparse
import os.path
import sys
from collections import Counter
from copy import deepcopy
from pathlib import Path
from typing import ByteString, Dict, List, Optional, Tuple, Type, Union
from uuid import uuid4

import attr
from cryptography.fernet import Fernet

from spock.backend.builder import AttrBuilder
from spock.backend.payload import AttrPayload
from spock.backend.resolvers import EnvResolver
from spock.backend.saver import AttrSaver
from spock.backend.wrappers import Spockspace
from spock.exceptions import _SpockCryptoError, _SpockEvolveError, _SpockValueError
from spock.handlers import YAMLHandler
from spock.utils import (
    _C,
    _T,
    _is_spock_instance,
    check_payload_overwrite,
    deep_payload_update,
    make_salt,
)


class ConfigArgBuilder:
    """Automatically generates dataclass instances from config file(s)

    This class builds out necessary arguments from *args classes, reads
    the arguments from specified config file(s), and subsequently (via chained
    call to generate) generates each class instance based on the necessary
    field values for each backend class instance

    Attributes:
        _args: all command line args
        _arg_namespace: generated argument namespace
        _builder_obj: instance of a BaseBuilder class
        _dict_args: dictionary args from the command line
        _payload_obj: instance of a BasePayload class
        _saver_obj: instance of a BaseSaver class
        _tune_payload_obj: payload for tuner related objects -- instance of TunerPayload class
        _tune_obj: instance of TunerBuilder class
        _tuner_interface: interface that handles the underlying library for sampling -- instance of TunerInterface
        _tuner_state: current state of the hyper-parameter sampler
        _tune_namespace: namespace that hold the generated tuner related parameters
        _sample_count: current call to the sample function
        _fixed_uuid: fixed uuid to write the best file to the same path
        _configs = configs if configs is None else [Path(c) for c in configs]
        _lazy: flag to lazily find @spock decorated classes registered within sys.modules["spock"].backend.config
            thus alleviating the need to pass all @spock decorated classes to *args
        _no_cmd_line: turn off cmd line args
        _desc: description for help
        _salt: salt use for crypto purposes
        _key: key used for crypto purposes

    """

    def __init__(
        self,
        *args,
        configs: Optional[List] = None,
        desc: str = "",
        lazy: bool = False,
        no_cmd_line: bool = False,
        s3_config: Optional[_T] = None,
        key: Optional[Union[str, ByteString]] = None,
        salt: Optional[str] = None,
        **kwargs,
    ):
        """Init call for ConfigArgBuilder

        Args:
            *args: tuple of spock decorated classes to process
            configs: list of config paths
            desc: description for help
            lazy: attempts to lazily find @spock decorated classes registered within sys.modules["spock"].backend.config
                as well as the parents of any lazily inherited @spock class thus alleviating the need to pass all
                @spock decorated classes to *args
            no_cmd_line: turn off cmd line args
            s3_config: s3Config object for S3 support
            salt: either a path to a prior spock saved salt.yaml file or a string of the salt (can be an env reference)
            key: either a path to a prior spock saved key.yaml file, a ByteString of the key, or a str of the key
                (can be an env reference)
            **kwargs: keyword args

        """
        # Do some verification first
        self._verify_attr(args)
        self._configs = configs if configs is None else [Path(c) for c in configs]
        self._lazy = lazy
        self._no_cmd_line = no_cmd_line
        self._desc = desc
        self._salt, self._key = self._maybe_crypto(key, salt, s3_config)
        # Build the payload and saver objects
        self._payload_obj = AttrPayload(s3_config=s3_config)
        self._saver_obj = AttrSaver(s3_config=s3_config)
        # Split the fixed parameters from the tuneable ones (if present)
        fixed_args, tune_args = self._strip_tune_parameters(args)
        # The fixed parameter builder
        self._builder_obj = AttrBuilder(
            *fixed_args, lazy=lazy, salt=self._salt, key=self._key, **kwargs
        )
        # The possible tunable parameter builder -- might return None
        self._tune_obj, self._tune_payload_obj = self._handle_tuner_objects(
            tune_args, s3_config, kwargs
        )
        self._tuner_interface = None
        self._tuner_state = None
        self._tuner_status = None
        self._sample_count = 0
        self._fixed_uuid = str(uuid4())
        try:
            # Get all cmd line args and build overrides
            self._args = self._handle_cmd_line()
            # Get the actual payload from the config files -- fixed configs
            self._dict_args = self._get_payload(
                payload_obj=self._payload_obj,
                input_classes=self._builder_obj.input_classes,
                ignore_args=tune_args,
            )
            # Build the Spockspace from the payload and the classes
            # Fixed configs
            self._arg_namespace = self._builder_obj.generate(self._dict_args)
            # Attach the key and salt to the Spockspace
            self._arg_namespace.__salt__ = self.salt
            self._arg_namespace.__key__ = self.key
            # Get the payload from the config files -- hyper-parameters -- only if the obj is not None
            if self._tune_obj is not None:
                self._tune_args = self._get_payload(
                    payload_obj=self._tune_payload_obj,
                    input_classes=self._tune_obj.input_classes,
                    ignore_args=fixed_args,
                )
                # Build the Spockspace from the payload and the classes
                # Tuneable parameters
                self._tune_namespace = self._tune_obj.generate(self._tune_args)
        except Exception as e:
            self._print_usage_and_exit(str(e), sys_exit=False)
            raise e

    def __call__(self, *args, **kwargs) -> _T:
        """Call to self to allow chaining

        Args:
            *args: non-keyword args
            **kwargs: keyword args

        Returns:
            ConfigArgBuilder: self instance
        """
        return ConfigArgBuilder(*args, **kwargs)

    def generate(self) -> Spockspace:
        """Generate method that returns the actual argument namespace


        Returns:
            argument namespace consisting of all config classes

        """
        return self._arg_namespace

    @property
    def tuner_status(self) -> Dict:
        """Returns a dictionary of all the necessary underlying tuner internals to report the result"""
        return self._tuner_status

    @property
    def best(self) -> Spockspace:
        """Returns a Spockspace of the best hyper-parameter config and the associated metric value"""
        return self._tuner_interface.best

    @property
    def salt(self):
        return self._salt

    @property
    def key(self):
        return self._key

    def sample(self) -> Spockspace:
        """Sample method that constructs a namespace from the fixed parameters and samples from the tuner space to
        generate a Spockspace derived from both

        Returns:
            argument namespace(s) -- fixed + drawn sample from tuner backend

        """
        if self._tuner_interface is None:
            raise RuntimeError(
                f"Called sample method without first calling the tuner method that initializes the "
                f"backend library"
            )
        return_tuple = self._tuner_state
        self._tuner_status = self._tuner_interface.tuner_status
        self._tuner_state = self._tuner_interface.sample()
        self._sample_count += 1
        return return_tuple

    def tuner(self, tuner_config: _T) -> _T:
        """Chained call that builds the tuner interface for either optuna or ax depending upon the type of the tuner_obj

        Args:
            tuner_config: a class of type optuna.study.Study or AX****

        Returns:
            self so that functions can be chained

        """

        if self._tune_obj is None:
            raise RuntimeError(
                f"Called tuner method without passing any @spockTuner decorated classes"
            )

        try:
            from spock.addons.tune.tuner import TunerInterface

            self._tuner_interface = TunerInterface(
                tuner_config=tuner_config,
                tuner_namespace=self._tune_namespace,
                fixed_namespace=self._arg_namespace,
            )
            self._tuner_state = self._tuner_interface.sample()
        except Exception as e:
            raise e
        return self

    def _print_usage_and_exit(
        self, msg: Optional[str] = None, sys_exit: bool = True, exit_code: int = 1
    ) -> None:
        """Prints the help message and exits

        Args:
            msg: message to print pre exit

        Returns:
            None

        """
        print(f"usage: {sys.argv[0]} -c [--config] config1 [config2, config3, ...]")
        print(f'\n{self._desc if self._desc != "" else ""}\n')
        print("configuration(s):\n")
        # Call the fixed parameter help info
        self._builder_obj.handle_help_info()
        if self._tune_obj is not None:
            self._tune_obj.handle_help_info()
        if msg is not None:
            print(msg)
        if sys_exit:
            sys.exit(exit_code)

    def _handle_tuner_objects(
        self, tune_args: List, s3_config: Optional[_T], kwargs: Dict
    ) -> Tuple:
        """Handles creating the tuner builder object if @spockTuner classes were passed in

        Args:
            tune_args: list of tuner classes
            s3_config: s3Config object for S3 support
            kwargs: optional keyword args

        Returns:
            tuner builder object or None

        """
        if len(tune_args) > 0:
            try:
                from spock.addons.tune.builder import TunerBuilder
                from spock.addons.tune.payload import TunerPayload

                tuner_builder = TunerBuilder(
                    *tune_args, **kwargs, lazy=self._lazy, salt=self.salt, key=self.key
                )
                tuner_payload = TunerPayload(s3_config=s3_config)
                return tuner_builder, tuner_payload
            except ImportError:
                print(
                    "Missing libraries to support tune functionality. Please re-install with the extra tune "
                    "dependencies -- pip install spock-config[tune]"
                )
        else:
            return None, None

    @staticmethod
    def _verify_attr(args: Tuple) -> None:
        """Verifies that all the input classes are attr based

        Args:
            args: tuple of classes passed to the builder

        Returns:
            None

        """
        # Gather if all attr backend
        type_attrs = all([attr.has(arg) for arg in args])
        if not type_attrs:
            which_idx = [attr.has(arg) for arg in args].index(False)
            if hasattr(args[which_idx], "__name__"):
                raise TypeError(
                    f"*args must be of all attrs backend -- missing a @spock decorator on class "
                    f"{args[which_idx].__name__}"
                )
            else:
                raise TypeError(
                    f"*args must be of all attrs backend -- invalid type "
                    f"{type(args[which_idx])}"
                )

    @staticmethod
    def _strip_tune_parameters(args: Tuple) -> Tuple[List, List]:
        """Separates the fixed arguments from any hyper-parameter arguments

        Args:
            args: tuple of classes passed to the builder

        Returns:
            fixed_args: list of fixed args
            tune_args: list of args destined for a tuner backend

        """
        fixed_args = []
        tune_args = []
        for arg in args:
            if arg.__module__ == "spock.backend.config":
                fixed_args.append(arg)
            elif arg.__module__ == "spock.addons.tune.config":
                tune_args.append(arg)
        return fixed_args, tune_args

    def _handle_cmd_line(self) -> argparse.Namespace:
        """Handle all cmd line related tasks

        Config paths can enter from either the command line or be added in the class init call
        as a kwarg (configs=[]) -- also trigger the building of the cmd line overrides for each fixed and
        tunable objects

        Returns:
            args: namespace of args

        """
        # Need to hold an overarching parser here that just gets appended to for both fixed and tunable objects
        # Check if the no_cmd_line is not flagged and if the configs are not empty
        if self._no_cmd_line and (self._configs is None):
            raise ValueError(
                "Flag set for preventing command line read but no paths were passed to the config kwarg"
            )
        # If cmd_line is flagged then build the parsers if not make any empty Namespace
        args = (
            self._build_override_parsers(desc=self._desc)
            if not self._no_cmd_line
            else argparse.Namespace(config=[], help=False)
        )
        # If configs are present from the init call then roll these into the namespace
        if self._configs is not None:
            args = self._get_from_kwargs(args, self._configs)
        return args

    def _build_override_parsers(self, desc: str) -> argparse.Namespace:
        """Creates parsers for command-line overrides

        Builds the basic command line parser for configs and help then iterates through each attr instance to make
        namespace specific cmd line override parsers -- handles calling both the fixed and tunable objects

        Args:
            desc: argparser description

        Returns:
            args: argument namespace

        """
        # Highest level parser object
        parser = argparse.ArgumentParser(description=desc, add_help=False)
        parser.add_argument(
            "-c", "--config", required=False, nargs="+", default=[], type=Path
        )
        parser.add_argument("-h", "--help", action="store_true")
        # Handle the builder obj
        parser = self._builder_obj.build_override_parsers(parser=parser)
        if self._tune_obj is not None:
            parser = self._tune_obj.build_override_parsers(parser=parser)
        args = parser.parse_args()
        return args

    @staticmethod
    def _get_from_kwargs(args: argparse.Namespace, configs: List):
        """Get configs from the configs kwarg

        Args:
            args: argument namespace
            configs: config kwarg

        Returns:
            args: arg namespace

        """
        if isinstance(configs, list):
            args.config.extend(configs)
        else:
            raise TypeError(
                f"configs kwarg must be of type list -- given {type(configs)}"
            )
        return args

    def _get_payload(
        self, payload_obj: Type[AttrPayload], input_classes: Tuple, ignore_args: List
    ) -> Dict:
        """Get the parameter payload from the config file(s)

        Calls the various ways to get configs and then parses to retrieve the parameter payload - make sure to call
        deep update so as to not lose some parameters when only partially updating the payload

        Args:
            payload_obj: current payload object to call
            input_classes: classes to use to get payload
            ignore_args: args that were decorated for hyper-parameter tuning

        Returns:
            payload: dictionary of parameter values

        """
        if self._args.help:
            # Call sys exit with a clean code as this is the help call which is not unexpected behavior
            self._print_usage_and_exit(sys_exit=True, exit_code=0)
        payload = {}
        dependencies = {"paths": [], "rel_paths": [], "roots": []}
        if payload_obj is not None:
            # Make sure we are actually trying to map to input classes
            if len(input_classes) > 0:
                # If configs are present then iterate through them and deal with the payload
                if len(self._args.config) > 0:
                    for configs in self._args.config:
                        payload_update = payload_obj.payload(
                            input_classes,
                            ignore_args,
                            configs,
                            self._args,
                            dependencies,
                        )
                        check_payload_overwrite(payload, payload_update, configs)
                        deep_payload_update(payload, payload_update)
                # If there are no configs present we have to fall back only on cmd line args to fill out the necessary
                # data -- this is essentially using spock as a drop in replacement of arg-parser
                else:
                    payload_update = payload_obj.payload(
                        input_classes, ignore_args, None, self._args, dependencies
                    )
                    check_payload_overwrite(payload, payload_update, None)
                    deep_payload_update(payload, payload_update)
        return payload

    def _save(
        self,
        payload: Spockspace,
        file_name: str = None,
        user_specified_path: Path = None,
        create_save_path: bool = True,
        extra_info: bool = True,
        file_extension: str = ".yaml",
        tuner_payload: Optional[Spockspace] = None,
        fixed_uuid: str = None,
    ) -> _T:
        """Private interface -- saves the current config setup to file with a UUID

        Args:
            payload: Spockspace to save
            file_name: name of file (will be appended with .spock.cfg.file_extension) -- falls back to just uuid if None
            user_specified_path: if user provides a path it will be used as the path to write
            create_save_path: bool to create the path to save if called
            extra_info: additional info to write to saved config (run date and git info)
            file_extension: file type to write (default: yaml)
            tuner_payload: tuner level payload (unsampled)
            fixed_uuid: fixed uuid to allow for file overwrite

        Returns:
            self so that functions can be chained
        """
        if user_specified_path is not None:
            save_path = Path(user_specified_path)
        elif self._builder_obj.save_path is not None:
            save_path = Path(self._builder_obj.save_path)
        else:
            raise ValueError(
                "Save did not receive a valid path from: (1) markup file(s) or (2) "
                "the keyword arg user_specified_path"
            )
        # Call the saver class and save function
        self._saver_obj.save(
            payload,
            save_path,
            file_name,
            create_save_path,
            extra_info,
            file_extension,
            tuner_payload,
            fixed_uuid,
        )
        return self

    def save(
        self,
        file_name: str = None,
        user_specified_path: Union[Path, str] = None,
        create_save_path: bool = True,
        extra_info: bool = True,
        file_extension: str = ".yaml",
        add_tuner_sample: bool = False,
    ) -> _T:
        """Saves the current config setup to file with a UUID

        Args:
            file_name: name of file (will be appended with .spock.cfg.file_extension) -- falls back to just uuid if None
            user_specified_path: if user provides a path it will be used as the path to write
            create_save_path: bool to create the path to save if called
            extra_info: additional info to write to saved config (run date and git info)
            file_extension: file type to write (default: yaml)
            add_tuner_sample: save the current tuner sample to the payload

        Returns:
            self so that functions can be chained
        """
        if user_specified_path is not None:
            user_specified_path = Path(user_specified_path)
        if add_tuner_sample:
            if self._tune_obj is None:
                raise ValueError(
                    f"Called save method with add_tuner_sample as `{add_tuner_sample}` without passing any @spockTuner "
                    f"decorated classes -- please use the add_tuner_sample flag for saving only hyper-parameter tuning "
                    f"runs"
                )
            file_name = (
                f"hp.sample.{self._sample_count+1}"
                if file_name is None
                else f"{file_name}.hp.sample.{self._sample_count+1}"
            )
            self._save(
                self._tuner_state,
                file_name,
                user_specified_path,
                create_save_path,
                extra_info,
                file_extension,
            )
        else:
            self._save(
                self._arg_namespace,
                file_name,
                user_specified_path,
                create_save_path,
                extra_info,
                file_extension,
                tuner_payload=self._tune_namespace
                if self._tune_obj is not None
                else None,
            )
        return self

    def save_best(
        self,
        file_name: str = None,
        user_specified_path: Path = None,
        create_save_path: bool = True,
        extra_info: bool = True,
        file_extension: str = ".yaml",
    ) -> _T:
        """Saves the current best config setup to file

        Args:
            file_name: name of file (will be appended with .spock.cfg.file_extension) -- falls back to just uuid if None
            user_specified_path: if user provides a path it will be used as the path to write
            create_save_path: bool to create the path to save if called
            extra_info: additional info to write to saved config (run date and git info)
            file_extension: file type to write (default: yaml)

        Returns:
            self so that functions can be chained
        """
        if self._tune_obj is None:
            raise ValueError(
                f"Called save_best method without passing any @spockTuner decorated classes -- please use the `save()`"
                f" method for saving non hyper-parameter tuning runs"
            )
        file_name = f"hp.best" if file_name is None else f"{file_name}.hp.best"
        self._save(
            Spockspace(**vars(self._arg_namespace), **vars(self.best[0])),
            file_name,
            user_specified_path,
            create_save_path,
            extra_info,
            file_extension,
            fixed_uuid=self._fixed_uuid,
        )

        return self

    @property
    def config_2_dict(self) -> Dict:
        """Dictionary representation of the arg payload"""
        return self._saver_obj.dict_payload(self._arg_namespace)

    def spockspace_2_dict(self, payload: Spockspace) -> Dict:
        """Converts an input SpockSpace into a dictionary

        Args:
            payload: SpockSpace generated by the ConfigArgBuilder

        Returns:
            dictionary representation of the SpockSpace

        """
        return self._saver_obj.dict_payload(payload)

    def obj_2_dict(self, obj: Union[_C, List[_C], Tuple[_C, ...]]) -> Dict[str, Dict]:
        """Converts spock classes from a Spockspace into their dictionary representations

        Args:
            objs: single spock class or an iterable of spock classes

        Returns:
            dictionary where the class names are keys and the values are the dictionary representations
        """
        if isinstance(obj, (List, Tuple)):
            obj_dict = {}
            for val in obj:
                if not _is_spock_instance(val):
                    raise _SpockValueError(
                        f"Object is not a @spock decorated class object -- currently `{type(val)}`"
                    )
                obj_dict.update({type(val).__name__: val})
        elif _is_spock_instance(obj):
            obj_dict = {type(obj).__name__: obj}
        else:
            raise _SpockValueError(
                f"Object is not a @spock decorated class object -- currently `{type(obj)}`"
            )
        return self.spockspace_2_dict(Spockspace(**obj_dict))

    def evolve(self, *args: _C) -> Spockspace:
        """Function that allows a user to evolve the underlying spock classes with instantiated spock objects

        This will map the differences between the passed in instantiated objects and the underlying class definitions
        to the underlying namespace -- this essentially allows you to 'evolve' the Spockspace similar to how attrs
        allows for class evolution -- returns a new Spockspace object

        Args:
            *args: variable number of instantiated @spock decorated classes to evolve parameters with

        Returns:
            new_arg_namespace: Spockspace evolved with *arg @spock decorated classes

        Raises:
            _SpockEvolveError: if multiple of the same instance are passed as input or if the one or more of the inputs
            are not within the set of original input classes

        """
        # First check that all instances are in the underlying set of input_classes and that there are no dupes
        arg_counts = Counter([type(v).__name__ for v in args])
        for k, v in arg_counts.items():
            if v > 1:
                raise _SpockEvolveError(
                    f"Passed multiple instances (count: {v}) of class `{k}` into `evolve()` -- please pass only a "
                    f"single instance of the class in order to evolve the underlying Spockspace"
                )
            elif k not in self._builder_obj.graph.node_names:
                raise _SpockEvolveError(
                    f"Passed class `{k}` into `evolve()` but that class in not within the set of input "
                    f"classes {repr(self._builder_obj.graph.node_names)}"
                )
        # Create a new copy of the object
        new_arg_namespace = deepcopy(self._arg_namespace)
        # Determine the order of overwrite ops -- based on the topological order
        topo_idx = sorted(
            zip(
                [
                    self._builder_obj.graph.topological_order.index(type(v).__name__)
                    for v in args
                ],
                args,
            )
        )
        args = {type(v).__name__: v for _, v in topo_idx}
        # Walk through the now sorted set of evolve classes
        for k, v in args.items():
            # Get the class name from the object
            cls_name = type(v).__name__
            # Swap in the value to the new object
            # Note: we don't need to evolve here as it's a fully new obj
            setattr(new_arg_namespace, cls_name, v)
            # Recurse upwards through the deps stack and evolve all the necessary classes
            new_arg_namespace, all_cls = self._recurse_upwards(
                new_arg_namespace, cls_name, args
            )
        return new_arg_namespace

    def _recurse_upwards(
        self, new_arg_namespace: Spockspace, current_cls: str, all_cls: Dict
    ) -> Tuple[Spockspace, Dict]:
        """Using the underlying graph work recurse upwards through the parents and swap in the correct values

        Args:
            new_arg_namespace: new Spockspace object
            current_cls: current name of the cls
            all_cls: dict of the variable number of @spock decorated classes to evolve parameters with

        Returns:
            modified new_arg_namespace and the updated evolve class dict

        """
        # Get the parent deps from the graph
        parents = self._builder_obj.dag[current_cls]
        if len(parents) > 0:
            for parent_cls in parents:
                parent_name = parent_cls.__name__
                # Change the parent classes in the Spockspace
                new_arg_namespace = self._set_matching_attrs_by_name(
                    new_arg_namespace, current_cls, parent_name
                )
                # if the parent is in the evolve classes then morph them too
                if parent_name in all_cls.keys():
                    all_cls = self._set_matching_attrs_by_name_args(
                        current_cls, parent_name, all_cls
                    )
                # then recurse to the parents
                new_arg_namespace, all_cls = self._recurse_upwards(
                    new_arg_namespace, parent_name, all_cls
                )
        return new_arg_namespace, all_cls

    @staticmethod
    def _set_matching_attrs_by_name_args(
        current_cls_name: str, parent_cls_name: str, all_cls: Dict
    ) -> Dict:
        """Sets the value of an attribute by matching it to a spock class name

        Args:
            current_cls_name: current name of the changed class
            parent_cls_name: name of the parent class that contains a reference to the current class
            all_cls: dict of the variable number of @spock decorated classes to evolve parameters with

        Returns:
            modified all_cls dictionary

        """
        new_arg_namespace = all_cls[parent_cls_name]
        names = attr.fields_dict(type(new_arg_namespace)).keys()
        for v in names:
            if type(getattr(new_arg_namespace, v)).__name__ == current_cls_name:
                # Some evolution magic -- attr library wants kwargs so trick it by unrolling the dict
                # This creates a new object
                new_obj = attr.evolve(
                    new_arg_namespace,
                    **{v: all_cls[current_cls_name]},
                )
                all_cls[parent_cls_name] = new_obj
                print(
                    f"Evolved CLS Dependency: Parent = {parent_cls_name}, Child = {current_cls_name}, Value = {v}"
                )
        return all_cls

    @staticmethod
    def _set_matching_attrs_by_name(
        new_arg_namespace: Spockspace, current_cls_name: str, parent_cls_name: str
    ) -> Spockspace:
        """Sets the value of an attribute by matching it to a spock class name

        Args:
            new_arg_namespace: new Spockspace object
            current_cls_name: current name of the changed class
            parent_cls_name: name of the parent class that contains a reference to the current class

        Returns:
            modified new_arg_namespace

        """
        parent_attr = getattr(new_arg_namespace, parent_cls_name)
        names = attr.fields_dict(type(parent_attr)).keys()
        for v in names:
            if type(getattr(parent_attr, v)).__name__ == current_cls_name:
                # Some evolution magic -- attr library wants kwargs so trick it by unrolling the dict
                # This creates a new object
                new_obj = attr.evolve(
                    getattr(new_arg_namespace, parent_cls_name),
                    **{v: getattr(new_arg_namespace, current_cls_name)},
                )
                # Swap the new object into the existing attribute slot
                setattr(new_arg_namespace, parent_cls_name, new_obj)
                print(
                    f"Evolved: Parent = {parent_cls_name}, Child = {current_cls_name}, Value = {v}"
                )
        return new_arg_namespace

    def _maybe_crypto(
        self,
        key: Optional[Union[str, ByteString]],
        salt: Optional[str],
        s3_config: Optional[_T] = None,
        salt_len: int = 16,
    ) -> Tuple[str, ByteString]:
        """Handles setting up the underlying cryptography needs

        Args:
            salt: either a path to a prior spock saved salt.yaml file or a string of the salt (can be an env reference)
            key: either a path to a prior spock saved key.yaml file, a ByteString of the key, or a str of the key
                (can be an env reference)
            s3_config: s3Config object for S3 support
            salt_len: length of the salt to create

        Returns:
            tuple containing a salt and a key that spock can use to hide parameters

        """
        env_resolver = EnvResolver()
        salt = self._get_salt(salt, env_resolver, salt_len, s3_config)
        key = self._get_key(key, env_resolver, s3_config)
        return salt, key

    def _get_salt(
        self,
        salt: Optional[str],
        env_resolver: EnvResolver,
        salt_len: int,
        s3_config: Optional[_T] = None,
    ) -> str:
        """

        Args:
            salt: either a path to a prior spock saved salt.yaml file or a string of the salt (can be an env reference)
            env_resolver: EnvResolver class to handle env variable resolution if needed
            salt_len: length of the salt to create
            s3_config: s3Config object for S3 support

        Returns:
            salt as a string

        """
        # Byte string is assumed to be a direct key
        if salt is None:
            salt = make_salt(salt_len)
        elif os.path.splitext(salt)[1] in {".yaml", ".YAML", ".yml", ".YML"}:
            salt = self._handle_yaml_read(salt, access="salt", s3_config=s3_config)
        else:
            salt, _ = env_resolver.resolve(salt, str)
        return salt

    def _get_key(
        self,
        key: Optional[Union[str, ByteString]],
        env_resolver: EnvResolver,
        s3_config: Optional[_T] = None,
    ) -> ByteString:
        """

        Args:
            key: either a path to a prior spock saved key.yaml file, a ByteString of the key, or a str of the key
                (can be an env reference)
            env_resolver: EnvResolver class to handle env variable resolution if needed
            s3_config: s3Config object for S3 support

        Returns:
            key as ByteString

        """
        if key is None:
            key = Fernet.generate_key()
        # Byte string is assumed to be a direct key
        elif os.path.splitext(key)[1] in {".yaml", ".YAML", ".yml", ".YML"}:
            key = self._handle_yaml_read(
                key, access="key", s3_config=s3_config, encode=True
            )
        else:
            # Byte string is assumed to be a direct key
            # So only handle the str here
            if isinstance(key, str):
                key, _ = env_resolver.resolve(key, str)
                key = str.encode(key)
        return key

    @staticmethod
    def _handle_yaml_read(
        value: str, access: str, s3_config: Optional[_T] = None, encode: bool = False
    ) -> Union[str, ByteString]:
        """Reads in a salt/key yaml

        Args:
            value: path to the key/salt yaml
            access: which variable name to use from the yaml
            s3_config: s3Config object for S3 support

        Returns:

        """
        # Read from the yaml and then split
        try:
            payload = YAMLHandler().load(Path(value), s3_config)
            read_value = payload[access]
            if encode:
                read_value = str.encode(read_value)
            return read_value
        except Exception as e:
            _SpockCryptoError(f"Attempted to read from path `{value}` but failed")



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""Creates the spock config interface that wraps attr -- tune version for hyper-parameters"""
import sys
from typing import List, Optional, Sequence, Tuple, Union
from uuid import uuid4

import attr
import optuna
from ax.modelbridge.generation_strategy import GenerationStrategy

from spock.backend.config import _base_attr


@attr.s(auto_attribs=True)
class AxTunerConfig:
    objective_name: str
    tracking_metric_names: Optional[List[str]] = None
    name: Optional[str] = f"spock_ax_{uuid4()}"
    minimize: bool = True
    parameter_constraints: Optional[List[str]] = None
    outcome_constraints: Optional[List[str]] = None
    support_intermediate_data: bool = False
    overwrite_existing_experiment: bool = False
    immutable_search_space_and_opt_config: bool = True
    is_test: bool = False
    generation_strategy: Optional[GenerationStrategy] = None
    enforce_sequential_optimization: bool = True
    random_seed: Optional[int] = None
    verbose_logging: bool = True


@attr.s(auto_attribs=True)
class OptunaTunerConfig:
    storage: Optional[Union[str, optuna.storages.BaseStorage]] = None
    sampler: Optional[optuna.samplers.BaseSampler] = None
    pruner: Optional[optuna.pruners.BasePruner] = None
    study_name: Optional[str] = f"spock_optuna_{uuid4()}"
    direction: Optional[Union[str, optuna.study.StudyDirection]] = None
    load_if_exists: bool = False
    directions: Optional[Sequence[Union[str, optuna.study.StudyDirection]]] = None


def _process_class(cls, kw_only: bool, make_init: bool, dynamic: bool):
    """Process a given class

    Args:
        cls: basic class definition
        kw_only: set kwarg only
        make_init: make an init function
        dynamic: allows inherited classes to not be @spock decorated

    Returns:
        cls with attrs dunder methods added

    """
    # Handles the MRO and gets old annotations
    bases, attrs_dict, merged_annotations = _base_attr(cls, kw_only, make_init, dynamic)
    # Dynamically make an attr class
    obj = attr.make_class(
        name=cls.__name__,
        bases=bases,
        attrs=attrs_dict,
        kw_only=kw_only,
        frozen=True,
        auto_attribs=True,
        init=make_init,
    )
    # For each class we dynamically create we need to register it within the system modules for pickle to work
    setattr(sys.modules["spock"].addons.tune.config, obj.__name__, obj)
    # Swap the __doc__ string from cls to obj
    obj.__doc__ = cls.__doc__
    # Set the __init__ function
    # Handle __annotations__ from the MRO
    obj.__annotations__ = merged_annotations
    return obj


def _spock_tune(
    maybe_cls=None,
    kw_only=True,
    make_init=True,
):
    """Ovverides basic spock_attr decorator with another name

    Using a different name allows spock to easily determine which parameters are normal and which are
    meant to be used in a hyper-parameter tuning backend

    Args:
        maybe_cls: maybe a basic class def maybe None depending on call type
        kw_only: Make all attributes keyword-only
        make_init: bool, define a __init__() method

    Returns:
        cls: attrs class that is frozen and kw only
    """

    def wrap(cls):
        return _process_class(cls, kw_only=kw_only, make_init=make_init, dynamic=False)

    # Note: Taken from dataclass/attr definition(s)
    # maybe_cls's type depends on the usage of the decorator.  It's a class
    # if it's used as `@spockTuner` but ``None`` if used as `@spockTuner()`.
    if maybe_cls is None:
        return wrap
    else:
        return wrap(maybe_cls)


# Make the alias for the decorator
spockTuner = _spock_tune


@attr.s
class RangeHyperParameter:
    """Range based hyper-parameter that is sampled uniformly

    Attributes:
        type: type of the hyper-parameter (note: spock will attempt to autocast into this type)
        bounds: min and max of the hyper-parameter range
        log_scale: log scale the values before sampling

    """

    type = attr.ib(
        type=str,
        validator=[
            attr.validators.instance_of(str),
            attr.validators.in_(["float", "int"]),
        ],
    )
    bounds = attr.ib(
        type=Union[Tuple[float, float], Tuple[int, int]],
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of((float, int)),
            iterable_validator=attr.validators.instance_of(tuple),
        ),
    )
    log_scale = attr.ib(type=bool, validator=attr.validators.instance_of(bool))


@attr.s
class ChoiceHyperParameter:
    """Choice based hyper-parameter that is sampled uniformly

    Attributes:
        type: type of the hyper-parameter -- (note: spock will attempt to autocast into this type)
        choices: list of variable length that contains all the possible choices to select from

    """

    type = attr.ib(
        type=str,
        validator=[
            attr.validators.instance_of(str),
            attr.validators.in_(["float", "int", "str", "bool"]),
        ],
    )
    choices = attr.ib(
        type=Union[List[str], List[int], List[float], List[bool]],
        validator=attr.validators.deep_iterable(
            member_validator=attr.validators.instance_of((float, int, bool, str)),
            iterable_validator=attr.validators.instance_of(list),
        ),
    )



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0


class _SpockUndecoratedClass(Exception):
    """Custom exception type for non spock decorated classes and not dynamic"""

    pass


class _SpockInstantiationError(Exception):
    """Custom exception for when the spock class cannot be instantiated correctly"""

    pass


class _SpockNotOptionalError(Exception):
    """Custom exception for missing value"""

    pass


class _SpockDuplicateArgumentError(Exception):
    """Custom exception type for duplicated values"""

    pass


class _SpockEvolveError(Exception):
    """Custom exception for when evolve errors occur"""

    pass


class _SpockValueError(Exception):
    """Custom exception for throwing value errors"""

    pass


class _SpockResolverError(Exception):
    """Custom exception for environment resolver"""

    pass


class _SpockCryptoError(Exception):
    """Custom exception for dealing with the crypto side of things"""

    pass



# -*- coding: utf-8 -*-

# Copyright FMR LLC <opensource@fidelity.com>
# SPDX-License-Identifier: Apache-2.0

"""
Spock is a framework that helps manage complex parameter configurations for Python applications

Please refer to the documentation provided in the README.md
"""

from spock.addons.s3.configs import S3Config, S3DownloadConfig, S3UploadConfig

__all__ = ["configs", "utils", "S3Config", "S3DownloadConfig", "S3UploadConfig"]
